import re, pyparsing, sympy, json
from typing import Reversible
from telnetlib import IP
from bionetgen.atomizer.utils.util import logMess
from bionetgen.atomizer.writer.bnglWriter import rindex

from sympy.printing.str import StrPrinter

prnter = StrPrinter({"full_prec": False})


class Parameter:
    def __init__(self):
        # spec is ID, value, constant or not, units
        self.Id = None
        self.val = None
        self.cts = None
        self.units = None
        self.rxn_ind = None

    def __str__(self):
        if self.units is not None and self.units != "":
            return "{} {} #{}".format(self.Id, self.val, self.units)
        else:
            return "{} {}".format(self.Id, self.val)

    def __repr__(self):
        return str(self)


class Compartment:
    def __init__(self):
        self.Id = None
        self.dim = None
        self.size = None
        self.cmt = None
        self.unit = None

    def __str__(self):
        if self.cmt is not None and self.cmt != "":
            txt = "{} {} {} #{}".format(self.Id, self.dim, self.size, self.cmt)
        else:
            txt = "{} {} {}".format(self.Id, self.dim, self.size)
        return txt

    def __repr__(self):
        return str(self)


class Molecule:
    def __init__(self):
        self.translator = {}
        self.initConc = -1
        self.initAmount = -1
        self.isConstant = False
        self.isBoundary = False
        self.compartment = None

    def __contains__(self, val):
        return val in str(self)

    def parse_raw(self, raw):
        self.raw = raw
        self.Id = raw["returnID"]
        self.initConc = raw["initialConcentration"]
        self.initAmount = raw["initialAmount"]
        self.isConstant = raw["isConstant"]
        self.isBoundary = raw["isBoundary"]
        self.compartment = raw["compartment"]
        self.name = raw["name"].replace(" ", "").replace("*", "m")
        self.identifier = raw["identifier"]
        self.conversionFactor = raw["conversionFactor"]

    def __str__(self):
        if self.Id in self.translator:
            # str2 is molecule types?
            txt = "{}".format(self.translator[self.Id].str2())
        else:
            txt = "{}()".format(self.Id)
        return txt

    def __repr__(self):
        return str(self)


class Species:
    def __init__(self):
        self.noCompartment = False
        self.translator = {}
        self.raw = None
        self.compartment = None
        self.Id = None
        self.initConc = None
        self.initAmount = None
        self.isConstant = None
        self.isBoundary = None
        self.isParam = False
        self.name = None
        self.identifier = None
        self.conversionFactor = None
        self.isConc = False
        self.concCorrected = False

    def parse_raw(self, raw):
        self.raw = raw
        self.Id = raw["returnID"]
        self.initConc = raw["initialConcentration"]
        self.initAmount = raw["initialAmount"]
        self.isConstant = raw["isConstant"]
        self.isBoundary = raw["isBoundary"]
        self.compartment = raw["compartment"]
        self.name = raw["name"].replace(" ", "")
        self.identifier = raw["identifier"]
        if self.initAmount >= 0:
            self.val = self.initAmount
        elif self.initConc >= 0:
            # TODO: Figure out what to do w/ conc
            self.isConc = True
            self.val = self.initConc
        else:
            self.val = 0
        self.conversionFactor = raw["conversionFactor"]

    def __str__(self):
        trans_id = (
            str(self.translator[self.Id])
            if self.Id in self.translator
            else self.Id + "()"
        )
        # handle conversion factor
        if self.conversionFactor is None:
            conv = ""
        else:
            conv = ""
            # conv = f"/{self.conversionFactor}"
        # decide if we are using a constant species
        mod = "$" if self.isConstant or self.isBoundary else ""
        if self.noCompartment or self.compartment == "" or self.compartment is None:
            if self.raw is not None:
                txt = "  {}{} {}{} #{} #{}".format(
                    mod,
                    trans_id,
                    self.val,
                    conv,
                    self.raw["returnID"],
                    self.raw["identifier"],
                )
            else:
                txt = "  {}{} {}{}".format(mod, trans_id, self.val, conv)
        else:
            # we have a compartment in our ID
            # need to make sure it's correct
            if "@" in trans_id:
                if re.search(r"(^@)", trans_id):
                    # @X: or @X:: syntax
                    if re.search(r"^@[\S\s]*::", trans_id):
                        trans_id = trans_id.split("::")[1]
                    else:
                        trans_id = trans_id.split(":")[1]
                else:
                    # X@Y syntax
                    trans_id = "@".join(trans_id.split("@")[:-1])
            # removing identical compartments because
            # we'll be using @comp: notation
            comp_str = "@{}".format(self.compartment)
            if comp_str in str(trans_id):
                trans_id = str(trans_id).replace(comp_str, "")
            if self.raw is not None:
                txt = "  @{}:{}{} {}{} #{} #{}".format(
                    self.compartment,
                    mod,
                    trans_id,
                    self.val,
                    conv,
                    self.raw["returnID"],
                    self.raw["identifier"],
                )
            else:
                txt = "  @{}:{}{} {}{}".format(
                    self.compartment, mod, trans_id, self.val, conv
                )
        return txt

    def __repr__(self):
        return str(self)


class Observable:
    def __init__(self):
        self.Id = None
        self.type = "Species"
        self.compartment = None
        self.noCompartment = False
        self.translator = {}
        self.raw = None
        self.pattern = None
        self.skip_compartment = False

    def parse_raw(self, raw):
        self.raw = raw
        self.Id = raw["returnID"]
        self.initConc = raw["initialConcentration"]
        self.initAmount = raw["initialAmount"]
        self.isConstant = raw["isConstant"]
        self.isBoundary = raw["isBoundary"]
        self.compartment = raw["compartment"]
        self.name = raw["name"].replace(" ", "")
        self.identifier = raw["identifier"]

    def get_obs_name(self):
        return self.Id
        # if self.skip_compartment:
        #     return self.Id
        # if self.noCompartment or self.compartment == "" or self.compartment is None:
        #     return self.Id
        # else:
        #     return "{0}_{1}".format(self.Id, self.compartment)

    def __str__(self):
        txt = self.type
        obs_name = self.get_obs_name()
        if self.pattern is None:
            if self.raw is not None:
                pattern = (
                    self.translator[self.raw["returnID"]]
                    if self.Id in self.translator
                    else self.raw["returnID"] + "()"
                )
            else:
                pattern = self.Id + "()"
        else:
            pattern = self.pattern
        if self.noCompartment or self.compartment == "":
            txt += " {0} {1} #{2}".format(obs_name, pattern, self.name)
        else:
            # removing identical compartments because
            # we'll be usgin @comp: notation
            comp_str = "@{}".format(self.compartment)
            if comp_str in str(pattern):
                pattern = str(pattern).replace(comp_str, "")
            txt += " {0} @{2}:{1} #{3}".format(
                obs_name, pattern, self.compartment, self.name
            )
        return txt

    def __repr__(self):
        return str(self)


class Function:
    def __init__(self):
        self.Id = None
        self.definition = None
        self.rule_ptr = None
        self.local_dict = None
        self.replaceLocParams = False
        self.all_syms = None
        self.sbmlFunctions = None
        self.compartmentList = None
        self.time_flag = False
        self.adjusted_def = None
        self.volume_adjusted = False

    def replaceLoc(self, func_def, pdict):
        if self.compartmentList is not None:
            if len(self.compartmentList) > 0:
                for comp in self.compartmentList:
                    cname, cval = comp
                    pdict[cname] = cval
        for parameter in pdict:
            func_def = re.sub(
                r"(\W|^)({0})(\W|$)".format(parameter),
                r"\g<1>{0}\g<3>".format(pdict[parameter]),
                func_def,
            )
        return func_def

    def renameLoc(self, pname, rind):
        return "r{}_{}".format(rind, pname)

    def __str__(self):
        fdef = self.definition
        if self.replaceLocParams:
            # check possible places, local dict first
            if self.local_dict is not None:
                fdef = self.replaceLoc(self.definition, self.local_dict)
            # or pull from the pointer to the rule itself
            elif self.rule_ptr is not None:
                if len(self.rule_ptr.raw_param) > 0:
                    rule_dict = dict([(i[0], i[1]) for i in self.rule_ptr.raw_param])
                    fdef = self.replaceLoc(self.definition, rule_dict)
        # if we are not replacing, we need to rename local parameters
        # to the correct index if the function is related to a rule
        else:
            if self.rule_ptr is not None:
                # this is a fRate, check for local parameters
                if len(self.rule_ptr.raw_param) > 0:
                    # gotta rename these in the function
                    rule_dict = dict(
                        [
                            (i[0], self.renameLoc(i[0], self.rule_ptr.rule_ind))
                            for i in self.rule_ptr.raw_param
                        ]
                    )
                    fdef = self.replaceLoc(self.definition, rule_dict)
        fdef = self.adjust_func_def(fdef)
        self.adjusted_def = fdef
        return "{} = {}".format(self.Id, fdef)

    def __repr__(self):
        return str(self)

    def adjust_func_def(self, fdef):
        # if this function is related to a rule, we'll pull all the
        # relevant info
        # TODO: Add sbml function resolution here
        if self.sbmlFunctions is not None:
            fdef = self.resolve_sbmlfuncs(fdef)

        if self.rule_ptr is not None:
            # TODO: pull info
            # react/prod/comp
            pass

        # This is stuff ported from bnglWriter
        # deals with comparison operators
        def compParse(match):
            translator = {
                "gt": ">",
                "lt": "<",
                "and": "&&",
                "or": "||",
                "geq": ">=",
                "leq": "<=",
                "eq": "==",
                "neq": "!=",
            }
            exponent = match.group(3)
            operator = translator[match.group(1)]
            return "{0} {1} {2}".format(match.group(2), operator, exponent)

        def changeToBNGL(functionList, rule, function):
            oldrule = ""
            # if the rule contains any mathematical function we need to reformat
            while any(
                [
                    re.search(r"(\W|^)({0})(\W|$)".format(x), rule) != None
                    for x in functionList
                ]
            ) and (oldrule != rule):
                oldrule = rule
                for x in functionList:
                    rule = re.sub("({0})\(([^,]+),([^)]+)\)".format(x), function, rule)
                if rule == oldrule:
                    logMess("ERROR:TRS001", "Malformed pow or root function %s" % rule)
            return rule

        def constructFromList(argList, optionList):
            parsedString = ""
            idx = 0
            translator = {
                "gt": ">",
                "lt": "<",
                "and": "&&",
                "or": "||",
                "geq": ">=",
                "leq": "<=",
                "eq": "==",
            }
            while idx < len(argList):
                if type(argList[idx]) is list:
                    parsedString += (
                        "(" + constructFromList(argList[idx], optionList) + ")"
                    )
                elif argList[idx] in optionList:
                    if argList[idx] == "ceil":
                        parsedString += "min(rint(({0}) + 0.5),rint(({0}) + 1))".format(
                            constructFromList(argList[idx + 1], optionList)
                        )
                        idx += 1
                    elif argList[idx] == "floor":
                        parsedString += (
                            "min(rint(({0}) -0.5),rint(({0}) + 0.5))".format(
                                constructFromList(argList[idx + 1], optionList)
                            )
                        )
                        idx += 1
                    elif argList[idx] in ["pow"]:
                        index = rindex(argList[idx + 1], ",")
                        parsedString += (
                            "(("
                            + constructFromList(argList[idx + 1][0:index], optionList)
                            + ")"
                        )
                        parsedString += (
                            " ^ "
                            + "("
                            + constructFromList(
                                argList[idx + 1][index + 1 :], optionList
                            )
                            + "))"
                        )
                        idx += 1
                    elif argList[idx] in ["sqr", "sqrt"]:
                        tag = "1/" if argList[idx] == "sqrt" else ""
                        parsedString += (
                            "(("
                            + constructFromList(argList[idx + 1], optionList)
                            + ") ^ ({0}2))".format(tag)
                        )
                        idx += 1
                    elif argList[idx] == "root":
                        index = rindex(argList[idx + 1], ",")
                        tmp = (
                            "1/("
                            + constructFromList(argList[idx + 1][0:index], optionList)
                            + "))"
                        )
                        parsedString += (
                            "(("
                            + constructFromList(
                                argList[idx + 1][index + 1 :], optionList
                            )
                            + ") ^ "
                            + tmp
                        )
                        idx += 1
                    elif argList[idx] == "piecewise":
                        index1 = argList[idx + 1].index(",")
                        try:
                            index2 = (
                                argList[idx + 1][index1 + 1 :].index(",") + index1 + 1
                            )
                            try:
                                index3 = (
                                    argList[idx + 1][index2 + 1 :].index(",")
                                    + index2
                                    + 1
                                )
                            except ValueError:
                                index3 = -1
                        except ValueError:
                            parsedString += constructFromList(
                                [argList[idx + 1][index1 + 1 :]], optionList
                            )
                            index2 = -1
                        if index2 != -1:
                            condition = constructFromList(
                                [argList[idx + 1][index1 + 1 : index2]], optionList
                            )
                            result = constructFromList(
                                [argList[idx + 1][:index1]], optionList
                            )
                            if index3 == -1:
                                result2 = constructFromList(
                                    [argList[idx + 1][index2 + 1 :]], optionList
                                )
                            else:
                                result2 = constructFromList(
                                    ["piecewise", argList[idx + 1][index2 + 1 :]],
                                    optionList,
                                )
                            parsedString += "if({0},{1},{2})".format(
                                condition, result, result2
                            )
                        idx += 1
                    elif argList[idx] in ["and", "or"]:
                        symbolDict = {"and": " && ", "or": " || "}
                        indexArray = [-1]
                        elementArray = []
                        for idx2, element in enumerate(argList[idx + 1]):
                            if element == ",":
                                indexArray.append(idx2)
                        indexArray.append(len(argList[idx + 1]))
                        tmpStr = argList[idx + 1]
                        for idx2, _ in enumerate(indexArray[0:-1]):
                            elementArray.append(
                                constructFromList(
                                    tmpStr[indexArray[idx2] + 1 : indexArray[idx2 + 1]],
                                    optionList,
                                )
                            )
                        parsedString += symbolDict[argList[idx]].join(elementArray)
                        idx += 1
                    elif argList[idx] == "lambda":
                        tmp = "("
                        try:
                            upperLimit = rindex(argList[idx + 1], ",")
                        except ValueError:
                            idx += 1
                            continue
                        parsedParams = []
                        for x in argList[idx + 1][0:upperLimit]:
                            if x == ",":
                                tmp += ", "
                            else:
                                tmp += "param_" + x
                                parsedParams.append(x)
                        tmp2 = ") = " + constructFromList(
                            argList[idx + 1][rindex(argList[idx + 1], ",") + 1 :],
                            optionList,
                        )
                        for x in parsedParams:
                            while (
                                re.search(r"(\W|^)({0})(\W|$)".format(x), tmp2) != None
                            ):
                                tmp2 = re.sub(
                                    r"(\W|^)({0})(\W|$)".format(x),
                                    r"\1param_\2 \3",
                                    tmp2,
                                )
                        idx += 1
                        parsedString += tmp + tmp2
                else:
                    parsedString += argList[idx]
                idx += 1
            return parsedString

        # This is where the changes happen
        # comparison operators sorted here
        fdef = changeToBNGL(["gt", "lt", "leq", "geq", "eq"], fdef, compParse)

        contentRule = (
            pyparsing.Word(pyparsing.alphanums + "_")
            | ","
            | "."
            | "+"
            | "-"
            | "*"
            | "/"
            | "^"
            | "&"
            | ">"
            | "<"
            | "="
            | "|"
        )
        parens = pyparsing.nestedExpr("(", ")", content=contentRule)
        finalString = ""

        if any(
            [
                re.search(r"(\W|^)({0})(\W|$)".format(x), fdef) != None
                for x in ["ceil", "floor", "pow", "sqrt", "sqr", "root", "and", "or"]
            ]
        ):
            argList = parens.parseString("(" + fdef + ")").asList()
            fdef = constructFromList(
                argList[0], ["floor", "ceil", "pow", "sqrt", "sqr", "root", "and", "or"]
            )

        while "piecewise" in fdef:
            argList = parens.parseString("(" + fdef + ")").asList()
            fdef = constructFromList(argList[0], ["piecewise"])
        # remove references to lambda functions
        if "lambda(" in fdef:
            lambdaList = parens.parseString("(" + fdef + ")")
            functionBody = constructFromList(lambdaList[0].asList(), ["lambda"])
            fdef = "{0}{1}".format(self.Id, functionBody)

        # change references to time for time()
        while re.search(r"(\W|^)inf(\W|$)", fdef) != None:
            fdef = re.sub(r"(\W|^)(inf)(\W|$)", r"\1 1e20 \3", fdef)

        # combinations '+ -' break bionetgen
        fdef = re.sub(r"(\W|^)([-])(\s)+", r"\1-", fdef)
        # pi
        fdef = re.sub(r"(\W|^)(pi)(\W|$)", r"\g<1>3.1415926535\g<3>", fdef)
        # log for log 10
        fdef = re.sub(r"(\W|^)log\(", r"\1 ln(", fdef)
        # reserved keyword: e
        fdef = re.sub(r"(\W|^)(e)(\W|$)", r"\g<1>__e__\g<3>", fdef)
        # TODO: Check if we need to replace local parameters
        # change references to local parameters
        # for parameter in parameterDict:
        #     finalString = re.sub(r'(\W|^)({0})(\W|$)'.format(parameter),r'\g<1>{0}\g<3>'.format(parameterDict[parameter]),finalString)
        # doing simplification
        try:
            sdef = sympy.sympify(fdef, locals=self.all_syms)
            fdef = prnter.doprint(sdef.nsimplify().evalf().simplify())
            fdef = fdef.replace("**", "^")
        except:
            pass
        return fdef

    def extendFunction(self, function, subfunctionName, subfunction):
        def constructFromList(argList, optionList, subfunctionParam, subfunctionBody):
            parsedString = ""
            idx = 0
            while idx < len(argList):
                if type(argList[idx]) is list:
                    parsedString += (
                        "("
                        + constructFromList(
                            argList[idx], optionList, subfunctionParam, subfunctionBody
                        )
                        + ")"
                    )
                elif argList[idx] in optionList:
                    tmp = subfunctionBody
                    commaIndexes = [0]
                    commaIndexes.extend(
                        [i for i, x in enumerate(argList[idx + 1]) if x == ","]
                    )
                    commaIndexes.append(len(argList[idx + 1]))
                    instancedParameters = [
                        argList[idx + 1][commaIndexes[i] : commaIndexes[i + 1]]
                        for i in range(0, len(commaIndexes) - 1)
                    ]
                    for parameter, instance in zip(
                        subfunctionParam, instancedParameters
                    ):
                        if "," in instance:
                            instance.remove(",")
                        parsedParameter = (
                            " ( "
                            + constructFromList(
                                instance, optionList, subfunctionParam, subfunctionBody
                            )
                            + " ) "
                        )
                        tmp = re.sub(
                            r"(\W|^)({0})(\W|$)".format(parameter.strip()),
                            r"\1{0} \3".format(parsedParameter),
                            tmp,
                        )
                    parsedString += " " + tmp + " "
                    idx += 1
                else:
                    if argList[idx] == "=":
                        parsedString += " " + argList[idx] + " "
                    else:
                        parsedString += argList[idx]
                idx += 1
            return parsedString

        param = subfunction.split(" = ")[0][len(subfunctionName) + 1 : -1]
        # ASS2019: There are cases where the fuction doesn't have a definition and the
        # following line errors out with IndexError, let's handle it.
        try:
            body = subfunction.split(" = ")[1]
        except IndexError as e:
            logMess(
                "ERROR:TRS002",
                "This function doesn't have a definition, note that atomizer doesn't allow for function linking: {}".format(
                    subfunction
                ),
            )
            raise e
        while (
            re.search(r"(\W|^){0}\([^)]*\)(\W|$)".format(subfunctionName), function)
            != None
        ):
            contentRule = (
                pyparsing.Word(pyparsing.alphanums + "_.")
                | ","
                | "+"
                | "-"
                | "*"
                | "/"
                | "^"
                | "&"
                | ">"
                | "<"
                | "="
                | "|"
            )
            parens = pyparsing.nestedExpr("(", ")", content=contentRule)
            subfunctionList = parens.parseString("(" + function + ")").asList()
            function = constructFromList(
                subfunctionList[0], [subfunctionName], param.split(","), body
            )
        return function

    def resolve_sbmlfuncs(self, defn):
        modificationFlag = True
        recursionIndex = 0
        # remove calls to other sbml functions
        while modificationFlag and recursionIndex < 20:
            modificationFlag = False
            for sbml in self.sbmlFunctions:
                if sbml in defn:
                    temp = self.extendFunction(defn, sbml, self.sbmlFunctions[sbml])
                    if temp != defn:
                        defn = temp
                        modificationFlag = True
                        recursionIndex += 1
                        break

        # AS-2021: temporary time fix until time function in
        # BioNetGen is fully fixed.
        # first check if we have time in the sbml function
        # time
        if re.search(r"(\W|^)(time)(\W|$)", defn):
            self.time_flag = True
            defn = re.sub(r"(\W|^)(time)(\W|$)", r"\1TIME_\3", defn)
        # Time
        if re.search(r"(\W|^)(Time)(\W|$)", defn):
            self.time_flag = True
            defn = re.sub(r"(\W|^)(Time)(\W|$)", r"\1TIME_\3", defn)
        # t
        if re.search(r"(\W|^)(t)(\W|$)", defn):
            self.time_flag = True
            defn = re.sub(r"(\W|^)(t)(\W|$)", r"\1TIME_\3", defn)

        # old code for the same purpose
        # defn = re.sub(r"(\W|^)(time)(\W|$)", r"\1time()\3", defn)
        # defn = re.sub(r"(\W|^)(Time)(\W|$)", r"\1time()\3", defn)
        # defn = re.sub(r"(\W|^)(t)(\W|$)", r"\1time()\3", defn)

        # remove true and false
        defn = re.sub(r"(\W|^)(true)(\W|$)", r"\1 1\3", defn)
        defn = re.sub(r"(\W|^)(false)(\W|$)", r"\1 0\3", defn)

        # TODO: Make sure we don't need these
        # dependencies2 = {}
        # for idx in range(0, len(functions)):
        #     dependencies2[functions[idx].split(' = ')[0].split('(')[0].strip()] = []
        #     for key in artificialObservables:
        #         oldfunc = functions[idx]
        #         functions[idx] = (re.sub(r'(\W|^)({0})([^\w(]|$)'.format(key), r'\1\2()\3', functions[idx]))
        #         if oldfunc != functions[idx]:
        #             dependencies2[functions[idx].split(' = ')[0].split('(')[0]].append(key)
        #     for element in sbmlfunctions:
        #         oldfunc = functions[idx]
        #         key = element.split(' = ')[0].split('(')[0]
        #         if re.search('(\W|^){0}(\W|$)'.format(key), functions[idx].split(' = ')[1]) != None:
        #             dependencies2[functions[idx].split(' = ')[0].split('(')[0]].append(key)
        #     for element in tfunc:
        #         key = element.split(' = ')[0].split('(')[0]
        #         if key in functions[idx].split(' = ')[1]:
        #             dependencies2[functions[idx].split( ' = ')[0].split('(')[0]].append(key)

        # fd = []
        # for function in functions:
        #     # print(function, '---', dependencies2[function.split(' = ' )[0].split('(')[0]], '---', function.split(' = ' )[0].split('(')[0], 0)
        #     fd.append([function, resolveDependencies(dependencies2, function.split(' = ' )[0].split('(')[0], 0)])
        # fd = sorted(fd, key= lambda rule:rule[1])
        # functions = [x[0] for x in fd]
        # return functions

        # returning expanded definition
        return defn


class Rule:
    def __init__(self):
        self.Id = ""
        self.reactants = []
        self.products = []
        self.rate_cts = (None,)
        self.comment = ""
        self.reversible = False
        self.translator = {}
        self.raw = None
        self.tags = None
        self.model = None
        self.raw_splt = False
        self.symm_factors = [1.0, 1.0]

    def parse_raw(self, raw):
        self.raw = raw
        self.raw_react = raw["reactants"]
        self.raw_prod = raw["products"]
        self.raw_param = raw["parameters"]
        self.raw_rates = raw["rates"]
        # self.raw_orates = raw['orates']
        self.raw_num = raw["numbers"]
        self.raw_splt = raw["split_rxn"]
        self.reversible = raw["reversible"]
        self.Id = raw["reactionID"]

    def __str__(self):
        if self.Id != "":
            txt = "{}: ".format(self.Id)
        else:
            txt = ""
        # reactants
        if len(self.reactants) == 0:
            txt += "0"
        else:
            for ir, react in enumerate(self.reactants):
                if ir != 0:
                    txt += " + "
                if react[0] in self.translator:
                    if self.tags is not None and not self.noCompartment:
                        if react[2] in self.tags:
                            tag = self.tags[react[2]]
                        elif react[0] in self.tags:
                            tag = self.tags[react[0]]
                        else:
                            tag = ""
                        react_str = str(self.translator[react[0]])
                        if not react_str.endswith(tag):
                            if "@" in tag:
                                splt = react_str.split("@")
                                tspl = tag.split("@")
                                if splt[-1] != tspl[-1]:
                                    splt[-1] = tspl[-1]
                                    react_str = "@".join(splt)
                    else:
                        react_str = str(self.translator[react[0]])
                else:
                    if self.tags is not None and not self.noCompartment:
                        if react[2] in self.tags:
                            react_str = str(react[0]) + "()" + self.tags[react[2]]
                        elif react[0] in self.tags:
                            react_str = str(react[0]) + "()" + self.tags[react[0]]
                        else:
                            react_str = str(react[0]) + "()"
                    else:
                        react_str = str(react[0]) + "()"
                # Apply stoichiometry
                # FIXME: What to do if stoichiometry is not an integer
                for i in range(int(react[1])):
                    if i > 0:
                        txt += " + "
                    txt += react_str
        # correct rxn arrow
        if self.reversible and len(self.rate_cts) == 2:
            txt += " <-> "
        else:
            txt += " -> "
        # products
        if len(self.products) == 0:
            txt += "0"
        else:
            for ip, prod in enumerate(self.products):
                if ip != 0:
                    txt += " + "
                if prod[0] in self.translator:
                    if self.tags is not None and not self.noCompartment:
                        if prod[2] in self.tags:
                            tag = self.tags[prod[2]]
                        elif react[0] in self.tags:
                            tag = self.tags[prod[0]]
                        else:
                            tag = ""
                        prod_str = str(self.translator[prod[0]])
                        if not prod_str.endswith(tag):
                            if "@" in tag:
                                splt = prod_str.split("@")
                                tspl = tag.split("@")
                                if splt[-1] != tspl[-1]:
                                    splt[-1] = tspl[-1]
                                    prod_str = "@".join(splt)
                    else:
                        prod_str = str(self.translator[prod[0]])
                else:
                    if self.tags is not None and not self.noCompartment:
                        if prod[2] in self.tags:
                            prod_str = str(prod[0]) + "()" + self.tags[prod[2]]
                        elif prod[0] in self.tags:
                            prod_str = str(prod[0]) + "()" + self.tags[prod[0]]
                        else:
                            prod_str = str(prod[0]) + "()"
                    else:
                        prod_str = str(prod[0]) + "()"
                # Apply stoichiometry
                # FIXME: What to do if stoichiometry is not an integer
                for i in range(int(prod[1])):
                    if i > 0:
                        txt += " + "
                    txt += prod_str
        if self.reversible and len(self.rate_cts) == 2:
            if self.model is not None:
                if len(self.model.param_repl) > 0:
                    for prep in self.model.param_repl:
                        if self.rate_cts[0] == prep:
                            self.rate_cts = (
                                self.model.param_repl[prep],
                                self.rate_cts[1],
                            )
                        if self.rate_cts[1] == prep:
                            self.rate_cts = (
                                self.rate_cts[0],
                                self.model.param_repl[prep],
                            )
        else:
            if self.model is not None:
                if len(self.model.param_repl) > 0:
                    for prep in self.model.param_repl:
                        if self.rate_cts[0] == prep:
                            self.rate_cts = (self.model.param_repl[prep],)
        if self.model is not None:
            # rate cts
            if self.reversible and len(self.rate_cts) == 2:
                # we need to check if the rate constant refers to an
                # observable and is alone
                if (
                    self.rate_cts[0] in self.model.obs_map
                    or self.rate_cts[0] in self.model.observables
                ):
                    r0 = "1*{0}".format(self.rate_cts[0])
                else:
                    r0 = "{}".format(self.rate_cts[0])
                if (
                    self.rate_cts[1] in self.model.obs_map
                    or self.rate_cts[1] in self.model.observables
                ):
                    r1 = "1*{0}".format(self.rate_cts[1])
                else:
                    r1 = "{}".format(self.rate_cts[1])
                txt += " {},{}".format(r0, r1)
            else:
                conv = ""
                if self.raw_splt:
                    # we can handle conversion factors here
                    # ensure we got the right type of 0 -> X rule
                    if len(self.reactants) == 0 and len(self.products) == 1:
                        prod = self.products[0]
                        if prod[0] in self.model.molecules:
                            if (
                                self.model.molecules[prod[0]].conversionFactor
                                is not None
                            ):
                                conv = (
                                    f"*{self.model.molecules[prod[0]].conversionFactor}"
                                )
                if (
                    self.rate_cts[0] in self.model.obs_map
                    or self.rate_cts[0] in self.model.observables
                ):
                    txt += " 1*{}{}".format(self.rate_cts[0], conv)
                else:

                    txt += " {}{}".format(self.rate_cts[0], conv)

        comment = ""
        if self.raw is not None:
            comment = (
                "Modifiers({0})".format(", ".join(self.raw["modifiers"]))
                if self.raw["modifiers"]
                else ""
            )
        if comment != "":
            txt += " #{}".format(comment)
        return txt

    def __repr__(self):
        return str(self)


class ARule:
    def __init__(self):
        self.Id = None
        self.rates = None
        self.isAssignment = None
        self.isRate = None

    def parse_raw(self, raw):
        self.Id = raw[0]
        self.rates = raw[1]
        self.isAssignment = raw[2]
        self.isRate = raw[3]

    def __str__(self):
        return "{} {}".format(self.Id, self.rates)

    def __repr__(self):
        return str(self)


# Model objects done


class bngModel:
    """
    Takes in atomizer stuff and turns everything
    into objects which can be used to print the
    final model
    """

    def __init__(self):
        self.parameters = {}
        self.compartments = {}
        self.molecules = {}
        self.species = {}
        self.observables = {}
        self.rules = {}
        self.arules = {}
        self.functions = {}
        #
        self.metaString = ""
        self.molecule_ids = {}
        self.translator = {}
        self.obs_map = {}
        self.molecule_mod_dict = {}
        self.parsed_func = {}
        self.noCompartment = None
        self.useID = False
        self.replaceLocParams = False
        self.all_syms = None
        self.function_order = None
        self.sbmlFunctions = None
        self.tags = None
        self.add_time = False
        self.param_repl = []
        self.obs_map_file = None
        self.used_in_rrule = []

    def __str__(self):
        txt = self.metaString

        txt += "begin model\n"

        if len(self.parameters.values()) > 0:
            txt += "begin parameters\n"
            for param in self.parameters.values():
                txt += "  " + str(param) + "\n"
            txt += "end parameters\n"

        if not self.noCompartment:
            txt += "begin compartments\n"
            for comp in self.compartments.values():
                txt += "  " + str(comp) + "\n"
            txt += "end compartments\n"

        if len(self.molecules.values()) > 0:
            txt += "begin molecule types\n"
            for molec in self.molecules.values():
                molec.translator = self.translator
                txt += "  " + str(molec) + "\n"
            txt += "end molecule types\n"

        if len(self.species.values()) > 0:
            txt += "begin seed species\n"
            for spec in self.species.values():
                spec.translator = self.translator
                if spec.Id in self.used_in_rrule:
                    spec.isBoundary = False
                if isinstance(spec.val, str):
                    spec.noCompartment = self.noCompartment
                    txt += f"{str(spec)}\n"
                elif spec.val > 0 or spec.isConstant or spec.isBoundary:
                    spec.noCompartment = self.noCompartment
                    txt += f"{str(spec)}\n"
            txt += "end seed species\n"

        if len(self.observables.values()) > 0:
            txt += "begin observables\n"
            for obs in self.observables.values():
                obs.translator = self.translator
                obs.noCompartment = self.noCompartment
                txt += "  " + str(obs) + "\n"
            txt += "end observables\n"

        if len(self.functions) > 0:
            txt += "begin functions\n"
            if self.function_order is None:
                for func in self.functions.values():
                    func.sbmlFunctions = self.sbmlFunctions
                    func.replaceLocParams = self.replaceLocParams
                    # we need to update the local dictionary
                    # with potential observable name changes
                    if len(self.obs_map) > 1:
                        if func.local_dict is not None:
                            func.local_dict.update(self.obs_map)
                        else:
                            func.local_dict = self.obs_map
                    if len(self.param_repl) > 1:
                        if func.local_dict is not None:
                            func.local_dict.update(self.param_repl)
                        else:
                            func.local_dict = self.param_repl
                    if func.Id in self.parsed_func:
                        func.sympy_parsed = self.parsed_func[func.Id]
                    func.all_syms = self.all_syms
                    txt += "  " + str(func) + "\n"
            else:
                for fkey in self.function_order:
                    func = self.functions[fkey]
                    func.sbmlFunctions = self.sbmlFunctions
                    func.replaceLocParams = self.replaceLocParams
                    # we need to update the local dictionary
                    # with potential observable name changes
                    if len(self.obs_map) > 1:
                        if func.local_dict is not None:
                            func.local_dict.update(self.obs_map)
                        else:
                            func.local_dict = self.obs_map
                    if len(self.param_repl) > 1:
                        if func.local_dict is not None:
                            func.local_dict.update(self.param_repl)
                        else:
                            func.local_dict = self.param_repl
                    if func.Id in self.parsed_func:
                        func.sympy_parsed = self.parsed_func[fkey]
                    func.all_syms = self.all_syms
                    txt += "  " + str(func) + "\n"
            txt += "end functions\n"

        if len(self.rules.values()) > 0:
            txt += "begin reaction rules\n"
            for rule in self.rules.values():
                rule.translator = self.translator
                rule.tags = self.tags
                rule.noCompartment = self.noCompartment
                rule.model = self
                txt += "  " + str(rule) + "\n"
            txt += "end reaction rules\n"

        txt += "end model"

        return txt

    def __repr__(self):
        return str((self.parameters, self.molecules))

    def _reset(self):
        self.molecules = {}
        self.molecule_ids = {}
        self.species = {}
        self.observables = {}

    def consolidate_arules(self):
        """
        this figures out what to do with particular
        assignment rules pulled from SBML.
        a) A non-constant parameter gets turned into
           a function
        b) Any species in the system can be modified
           by an assignment rule. This turns the species
           into a function which also requires a modification
           of any reaction rules the species is associated with
        c) rate rules get turned into syn reactions
        """
        for arule in self.arules.values():
            # first one is to check parameters
            if arule.isRate:
                # this is a rate rule, it'll be turned into a reaction
                # first make the entry in molecules
                if len(self.compartments) > 0 and not self.noCompartment:
                    comp = list(self.compartments.values())[0].Id
                else:
                    comp = None
                amolec = self.make_molecule()
                amolec.Id = arule.Id
                amolec.name = arule.Id
                if comp is not None:
                    amolec.compartment = self.compartments[comp]
                self.add_molecule(amolec)
                # turn the rate cts into a function
                nfunc = self.make_function()
                nfunc.Id = "rrate_{}".format(amolec.Id)
                # we need to divide by volume if we have a compartment
                if comp is not None:
                    # we also need to check that the definition actually has
                    # species that reside in a volume
                    nfunc.definition = arule.rates[0]
                    corrected = False
                    if not nfunc.volume_adjusted:
                        for mid in self.molecule_ids:
                            if mid in arule.rates[0]:
                                vol = self.compartments[comp].size
                                nfunc.definition = nfunc.definition.replace(
                                    mid, f"({mid})/{vol}"
                                )
                                corrected = True
                    nfunc.volume_adjusted = corrected
                else:
                    nfunc.definition = arule.rates[0]
                self.add_function(nfunc)
                # now make the rule
                if comp is not None:
                    prod_id = "{}()@{}".format(arule.Id, comp)
                else:
                    prod_id = "{}".format(arule.Id)
                nrule = self.make_rule()
                nrule.Id = "rrule_{}".format(arule.Id)
                nrule.products.append([prod_id, 1.0, prod_id])
                nrule.rate_cts = (nfunc.Id,)
                self.add_rule(nrule)
                # add observable
                nobs = self.make_observable()
                nobs.Id = arule.Id
                nobs.name = "rrule_{}".format(arule.Id)
                nobs.compartment = comp
                self.add_observable(nobs)
                # remove from parameters if exists
                # otherwise we can get namespace clashes
                # with observables
                if arule.Id in self.parameters:
                    seed_val = self.parameters.pop(arule.Id).val
                else:
                    seed_val = 0
                # add species
                nspec = self.make_species()
                nspec.Id = arule.Id
                nspec.name = arule.Id
                nspec.val = seed_val
                nspec.isConstant = False
                if comp is not None:
                    nspec.compartment = comp
                self.add_species(nspec)
                self.used_in_rrule.append(nspec.Id)
            elif arule.isAssignment:
                # rule is an assignment rule
                # let's first check parameters
                if arule.Id in self.parameters:
                    a_param = self.parameters[arule.Id]
                    # if not a_param.cts:
                    # this means that one of our parameters
                    # is _not_ a constant and is modified by
                    # an assignment rule
                    # TODO: Not sure if anything else
                    # can happen here. Confirm via SBML spec
                    a_param = self.parameters.pop(arule.Id)
                    # TODO: check if an initial value to
                    # a non-constant parameter is relevant?
                    # I think the only thing we need is to
                    # turn this into a function
                    fobj = self.make_function()
                    fobj.Id = arule.Id
                    fobj.definition = arule.rates[0]
                    self.add_function(fobj)
                elif arule.Id in self.molecule_ids:
                    # we are an assignment rule that modifies
                    # a molecule, this will be converted to
                    # a function if true
                    mname = self.molecule_ids[arule.Id]
                    molec = self.molecules[mname]
                    # We can't have the molecule be _constant_
                    # at which point it's supposed to be encoded
                    # with "$" in BNGL
                    if not molec.isConstant:
                        # we can have it be boundary or not, doesn't
                        # matter since we know an assignment rule is
                        # modifying it and it will take over reactions

                        # this should be guaranteed
                        molec = self.molecules.pop(mname)

                        # we should also remove this from species
                        # and/or observables, this checks for
                        # namespace collisions.
                        # TODO: We might want to
                        # remove parameters as well
                        if molec.name in self.observables:
                            obs = self.observables.pop(molec.name)
                            self.obs_map[obs.get_obs_name()] = molec.Id + "()"
                        elif molec.Id in self.observables:
                            obs = self.observables.pop(molec.Id)
                            self.obs_map[obs.get_obs_name()] = molec.Id + "()"
                        # for spec in self.species:
                        #     sobj = self.species[spec]
                        #     # if molec.name == sobj.Id or molec
                        if molec.name in self.species:
                            spec = self.species.pop(molec.name)
                        elif molec.Id in self.species:
                            spec = self.species.pop(molec.Id)
                        if molec.Id in self.parameters:
                            param = self.parameters.pop(molec.Id)

                        # this will be a function
                        fobj = self.make_function()
                        # TODO: sometimes molec.name is not
                        # normalized, check if .Id works consistently
                        fobj.Id = molec.Id + "()"
                        fobj.definition = arule.rates[0]
                        if len(arule.compartmentList) > 0:
                            fobj.local_dict = {}
                            for comp in arule.compartmentList:
                                cname, cval = comp
                                fobj.local_dict[cname] = cval
                        self.add_function(fobj)
                        # we want to make sure arules are the only
                        # things that change species concentrations
                        if (
                            mname in self.molecule_mod_dict
                            or molec.Id in self.molecule_mod_dict
                        ):
                            if mname in self.molecule_mod_dict:
                                mkey = mname
                            else:
                                mkey = molec.Id
                            for rule in self.molecule_mod_dict[mkey]:
                                if len(rule.reactants) == 0 and len(rule.products) == 1:
                                    # this is a syn rule, should be only generating the species in question
                                    if mkey == rule.products[0][0]:
                                        if rule.Id in self.rules:
                                            self.rules.pop(rule.Id)
                                else:
                                    # this is a more complicated rule, we need to adjust the rates
                                    for ir, react in enumerate(rule.reactants):
                                        if react[0] == mkey:
                                            # we have the molecule in reactants
                                            if len(rule.rate_cts) == 2:
                                                r = rule.reactants.pop(ir)
                                                fw, bk = rule.rate_cts
                                                rule.rate_cts = (
                                                    "{0}*".format(mkey) + fw,
                                                    bk,
                                                )
                                            else:
                                                r = rule.reactants.pop(ir)
                                                fw = rule.rate_cts[0]
                                                rule.rate_cts = (
                                                    "{0}*".format(mkey) + fw,
                                                )
                                    for ip, prod in enumerate(rule.products):
                                        if prod[0] == mkey:
                                            # molecule in products
                                            if len(rule.rate_cts) == 2:
                                                # adjust back rate
                                                p = rule.products.pop(ip)
                                                fw, bk = rule.rate_cts
                                                rule.rate_cts = (
                                                    fw,
                                                    "{0}*".format(mkey) + bk,
                                                )
                                            else:
                                                # we can just remove
                                                rule.products.pop(ip)
                                    if len(rule.reactants) == 0 and len(rule.products):
                                        if rule.Id in self.rules:
                                            self.rules.pop(rule.Id)

                else:
                    # this is just a simple assignment (hopefully)
                    # just convert to a function
                    fobj = self.make_function()
                    fobj.Id = arule.Id + "()"
                    fobj.definition = arule.rates[0]
                    self.add_function(fobj)
                    # we also might need to remove these from
                    # observables
                    if arule.Id in self.observables:
                        obs = self.observables.pop(arule.Id)
                        self.obs_map[obs.get_obs_name()] = fobj.Id
                    # we also have to remove this from rules
                    if arule.Id in self.molecule_mod_dict:
                        mkey = arule.Id
                        for rule in self.molecule_mod_dict[mkey]:
                            if len(rule.reactants) == 0 and len(rule.products) == 1:
                                # this is a syn rule, should be only generating the species in question
                                if mkey == rule.products[0][0]:
                                    if rule.Id in self.rules:
                                        self.rules.pop(rule.Id)
                            else:
                                # this is a more complicated rule, we need to adjust the rates
                                for ir, react in enumerate(rule.reactants):
                                    if react[0] == mkey:
                                        # we have the molecule in reactants
                                        if len(rule.rate_cts) == 2:
                                            r = rule.reactants.pop(ir)
                                            fw, bk = rule.rate_cts
                                            rule.rate_cts = (
                                                "{0}*".format(mkey) + fw,
                                                bk,
                                            )
                                        else:
                                            r = rule.reactants.pop(ir)
                                            fw = rule.rate_cts[0]
                                            rule.rate_cts = ("{0}*".format(mkey) + fw,)
                                for ip, prod in enumerate(rule.products):
                                    if prod[0] == mkey:
                                        # molecule in products
                                        if len(rule.rate_cts) == 2:
                                            # adjust back rate
                                            p = rule.products.pop(ip)
                                            fw, bk = rule.rate_cts
                                            rule.rate_cts = (
                                                fw,
                                                "{0}*".format(mkey) + bk,
                                            )
                                        else:
                                            # we can just remove
                                            rule.products.pop(ip)
                                if len(rule.reactants) == 0 and len(rule.products):
                                    if rule.Id in self.rules:
                                        self.rules.pop(rule.Id)
            else:
                # not sure what this means, read SBML spec more
                pass

    def consolidate_molecules(self):
        # potentially remove unused ones
        # or EmptySet and similar useless ones
        turn_param = []
        str_comp = []
        to_remove = []
        for molec in self.molecules:
            if molec not in self.molecule_mod_dict:
                if self.molecules[molec].isConstant:
                    if not self.molecules[molec].isBoundary:
                        turn_param.append(molec)
                        continue
            mstr = str(self.molecules[molec])
            if mstr not in str_comp:
                str_comp.append(mstr)
            else:
                # we already have this
                to_remove.append(molec)
        for torem in to_remove:
            self.molecules.pop(torem)
        for molec in turn_param:
            m = self.molecules.pop(molec)
            param = self.make_parameter()
            param.Id = m.Id
            # potentially we have some molecules set to parameters
            # already from atomizer, check for that first
            if m.name in self.species:
                # check if we already got the param decision from atomizer
                sp = self.species[m.name]
                if sp.isParam:
                    # this means we already have a decision
                    param.val = sp.val
                else:
                    param.val = m.initConc if m.initConc > 0 else m.initAmount
            else:
                param.val = m.initConc if m.initConc > 0 else m.initAmount
            self.add_parameter(param)
            if m.name in self.observables:
                self.observables.pop(m.name)
            elif m.Id in self.observables:
                self.observables.pop(m.Id)
            if m.name in self.species:
                self.species.pop(m.name)
            elif m.Id in self.species:
                self.species.pop(m.Id)

    def consolidate_observables(self):
        # if we are using compartments, we need
        # to adjust names in functions to match
        # with new observable names
        for obs in self.observables:
            obs_obj = self.observables[obs]
            oname = obs_obj.get_obs_name()
            self.obs_map[obs_obj.Id] = oname
            if oname in self.parameters:
                self.parameters.pop(oname)

    def consolidate_compartments(self):
        if len(self.compartments) == 1:
            comp_key = list(self.compartments.keys())[0]
            comp = self.compartments[comp_key]
            if comp.size == 1.0:
                _ = self.compartments.pop(comp_key)
                self.noCompartment = True

    def consolidate(self):
        self.consolidate_compartments()
        self.consolidate_arules()
        self.adjust_frate_functions()
        self.consolidate_molecules()
        self.consolidate_observables()
        self.reorder_functions()
        str(self)
        self.check_for_time_function()
        # self.adjust_volume_corrections()
        self.adjust_concentrations()
        self.print_obs_map()

    def adjust_concentrations(self):
        # some species are given as concentrations
        # we need to convert them to amounts
        for spec in self.species:
            s = self.species[spec]
            if s.isConc:
                if s.compartment in self.compartments:
                    comp = self.compartments[s.compartment]
                    s.val = s.initConc * comp.size
                    s.concCorrected = True
                    s.isConc = False

    def adjust_volume_corrections(self):
        if self.noCompartment:
            return
        for rname in self.rules:
            rule = self.rules[rname]
            if (not rule.reversible) and (len(rule.reactants) > 1):
                react_set = set([react[0] for react in rule.reactants])
                if len(react_set) > 1:
                    # we have more than one type of reactant
                    # we need to see if the rate law is a constant
                    # and if so, we can adjust the volume correction
                    # done by bionetgen by multiplying by the volume
                    if rule.rate_cts[0] in self.parameters:
                        # first pass test to see if this is a single constant
                        # now we need the compartment volume
                        # FIXME: what do we do if we have more than one compartment?
                        react_names = [react[0] for react in rule.reactants]
                        correction = False
                        for react_name in react_names:
                            if correction:
                                break
                            if react_name in rule.tags:
                                if "@" in rule.tags[react_name]:
                                    comp_name = rule.tags[react_name].replace("@", "")
                                    if comp_name in self.compartments:
                                        comp = self.compartments[comp_name]
                                        vol = comp.size
                                        rule.rate_cts = (f"({rule.rate_cts[0]})*{vol}",)
                                        correction = True
                                        break
            elif rule.reversible and (len(rule.reactants) > 1):
                # we don't know what's going on with reversible reactions right now
                pass

    def adjust_frate_functions(self):
        if not hasattr(self, "compartments"):
            return
        if len(self.compartments) == 0:
            return
        for rule_name in self.rules:
            rule = self.rules[rule_name]
            # if rule.raw_splt:
            # we are a split reaction and likely have fRate as our rate constant
            if "fRate" in rule.rate_cts[0]:
                # we got the fRate in the definition, let's get the value
                frate_search = re.search("fRate.+\(\)", rule.rate_cts[0])
                if frate_search:
                    frate_name = frate_search.group(0)
                    # we got the name
                    frate = self.functions[frate_name]
                    if not frate.volume_adjusted:
                        corrected = False
                        for spec_name in self.species:
                            # if frate.volume_adjusted:
                            #     break
                            if spec_name in frate.definition:
                                # means we got a volume to divide by
                                # TODO: Wtf happens if this has multiple species?
                                sp = self.species[spec_name]
                                comp = self.compartments[sp.compartment]
                                vol = comp.size
                                frate.definition = frate.definition.replace(
                                    spec_name, f"({spec_name}/{vol})"
                                )
                                # frate.volume_adjusted = True
                                # break
                                corrected = True
                        frate.volume_adjusted = corrected
                else:
                    print(f"we couldn't find frate in {rule.rate_cts[0]}")

    def print_obs_map(self):
        if self.obs_map_file is not None:
            obs_map = {}
            for obs in self.observables:
                obs_obj = self.observables[obs]
                if obs_obj.name not in obs_map:
                    obs_map[obs_obj.name] = obs_obj.Id
            with open(self.obs_map_file, "w") as f:
                json.dump(obs_map, f)

    def check_for_time_function(self):
        # see if SBML functions refer to time() in bngl
        # and if so add replace with an observable and
        # make a time counter reaction
        for func in self.functions.values():
            # time
            if re.search(r"(\W|^)(TIME_)(\W|$)", func.adjusted_def):
                self.add_time = True
            if self.add_time:
                break

        if self.add_time:
            self.add_time_rule()

    def add_time_rule(self):
        if len(self.compartments) > 0 and not self.noCompartment:
            comp = list(self.compartments.values())[0].Id
        else:
            comp = None
        # add molecule
        amolec = self.make_molecule()
        amolec.Id = "time_"
        amolec.name = "time_"
        if comp is not None:
            amolec.compartment = self.compartments[comp]
        self.add_molecule(amolec)

        # add observable
        nobs = self.make_observable()
        nobs.skip_compartment = True
        nobs.type = "Molecules"
        nobs.Id = "TIME_"
        nobs.name = "TIME_"
        nobs.pattern = "time_()"

        if comp is not None:
            nobs.compartment = comp
        self.add_observable(nobs)

        # self.make_rule
        nrule = self.make_rule()
        nrule.Id = "time_rule"
        if comp is not None:
            prod = f"@{comp}:{amolec.name}"
        else:
            prod = f"{amolec.name}"
        nrule.products.append([prod, 1.0, prod])
        nrule.rate_cts = ("1",)
        self.add_rule(nrule)

    def remove_sympy_symbols(self, fdef):
        to_replace = {
            "def": "__DEF__",
            "lambda": "__LAMBDA__",
            "as": "__AS__",
            "del": "__DEL__",
        }
        for rep_str in to_replace:
            if re.search(r"(\W|^)({0})(\W|$)".format(rep_str), fdef):
                fdef = re.sub(
                    r"(\W|^)({0})(\W|$)".format(rep_str),
                    r"\g<1>{0}\g<3>".format(to_replace[rep_str]),
                    fdef,
                )
        return fdef

    def reorder_functions(self):
        """
        this one is to make sure the functions are reordered
        correctly, should be ported from the original codebase
        """
        # initialize dependency graph
        func_names = {}
        dep_dict = {}
        for fkey in self.functions:
            func = self.functions[fkey]
            # this finds the pure function name
            # with or without parans
            ma = re.search(r"(\W|^)(\w*)(\(*)(\w*)(\)*)(\W|$)", fkey)
            pure_name = ma.group(2)
            func_names[pure_name] = func.Id
            if "fRate" not in func.Id:
                dep_dict[func.Id] = []
        # make dependency graph between funcs only
        func_order = []
        unparsed = []
        frates = []
        func_dict = {}
        # Let's replace and build dependency map
        for fkey in self.functions:
            func = self.functions[fkey]
            f = func.definition
            f = self.remove_sympy_symbols(f)
            try:
                fs = sympy.sympify(f, locals=self.all_syms)
                self.parsed_func[fkey] = fs
            except:
                # Can't parse this func
                if fkey.startswith("fRate"):
                    frates.append(fkey)
                else:
                    unparsed.append(fkey)
                continue
            func_dict[fkey] = fs
            # need to build a dependency graph to figure out what to
            # write first
            # We can skip this if it's a functionRate
            if "fRate" not in fkey:
                list_of_deps = list(map(str, fs.atoms(sympy.Symbol)))
                for dep in list_of_deps:
                    if dep in func_names:
                        dep_dict[fkey].append(func_names[dep])
            else:
                frates.append(fkey)
        # Now reorder accordingly
        ordered_funcs = []
        # this ensures we write the independendent functions first
        stck = sorted(dep_dict.keys(), key=lambda x: len(dep_dict[x]))
        # FIXME: This algorithm works but likely inefficient
        while len(stck) > 0:
            k = stck.pop()
            deps = dep_dict[k]
            if len(deps) == 0:
                if k not in ordered_funcs:
                    ordered_funcs.append(k)
            else:
                stck.append(k)
                for dep in deps:
                    if dep not in ordered_funcs:
                        stck.append(dep)
                    dep_dict[k].remove(dep)
        # print ordered functions and return
        ordered_funcs += frates
        self.function_order = ordered_funcs

    # model object creator and adder methods
    def add_parameter(self, param):
        self.parameters[param.Id] = param

    def make_parameter(self):
        return Parameter()

    def add_compartment(self, comp):
        # TODO: check if we really want this, this
        # replaces compartment in functions with their size
        self.obs_map[comp.Id] = comp.size
        self.compartments[comp.Id] = comp

    def make_compartment(self):
        return Compartment()

    def add_molecule(self, molec):
        # we might have to add molecules that
        # didn't have rawSpecies associated with
        if hasattr(molec, "raw"):
            self.molecule_ids[molec.raw["identifier"]] = molec.name
        if not molec.name in self.molecules:
            self.molecules[molec.name] = molec
        else:
            # TODO: check if this actually works for
            # everything, there are some cases where
            # the same molecule is actually different
            # e.g. 103
            if not molec.Id in self.molecules:
                self.molecules[molec.Id] = molec
            elif hasattr(molec, "raw"):
                self.molecules[molec.identifier] = molec
            else:
                print("molecule doesn't have identifier {}".format(molec))
                pass

    def make_molecule(self):
        return Molecule()

    def add_species(self, sspec):
        if not sspec.Id in self.species:
            self.species[sspec.Id] = sspec
        elif hasattr(sspec, "identifier"):
            self.species[sspec.identifier] = sspec
        else:
            print("species doesn't have identifier {}".format(sspec))
            pass

    def make_species(self):
        return Species()

    def add_observable(self, obs):
        if not obs.Id in self.observables:
            self.observables[obs.Id] = obs
        elif hasattr(obs, "identifier"):
            self.observables[obs.identifier] = obs
        else:
            print("observable doesn't have identifier {}".format(obs))
            pass

    def make_observable(self):
        return Observable()

    def make_function(self):
        return Function()

    def add_function(self, func):
        self.functions[func.Id] = func

    def make_rule(self):
        return Rule()

    def add_rule(self, rule):
        # add this to keep track of molecule modifications
        # this will allow us to quickly loop over reactons
        # a molecule is a part of if the molecule gets
        # turned to a function
        for react in rule.reactants:
            if react[0] not in self.molecule_mod_dict:
                self.molecule_mod_dict[react[0]] = []
            self.molecule_mod_dict[react[0]].append(rule)
        for prod in rule.products:
            if prod[0] not in self.molecule_mod_dict:
                self.molecule_mod_dict[prod[0]] = []
            self.molecule_mod_dict[prod[0]].append(rule)
        self.rules[rule.Id] = rule

    def make_arule(self):
        return ARule()

    def add_arule(self, arule):
        self.arules[arule.Id] = arule



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 12 13:44:53 2013

@author: proto
"""

import csv

try:
    import pandas as pd
except ImportError:
    pd = None


def loadBioGrid(fileName="BIOGRID-ALL-3.2.108.tab2.txt"):
    if hasattr(loadBioGrid, "db"):
        return loadBioGrid.db
    loadBioGrid.db = {}

    def check(db, term):
        if term not in db:
            db[term] = set([])

    f = open(fileName, "rb")
    content = csv.DictReader(f, delimiter="\t")
    loadBioGrid.db["synonyms"] = {}

    for idx, line in enumerate(content):
        tmp = {}
        tmp["osiB"] = line["Official Symbol Interactor B"]
        tmp["osiA"] = line["Official Symbol Interactor A"]
        tmp["siA"] = line["Synonyms Interactor A"].split("|")
        tmp["siB"] = line["Synonyms Interactor B"].split("|")

        tmp["siA"].append(tmp["osiA"])
        tmp["siB"].append(tmp["osiB"])
        comb = ([x, y] for x in tmp["siA"] for y in tmp["siB"])

        """
        for element in tmp['siA']:
            check(loadBioGrid.db['synonyms'], element.upper())
            loadBioGrid.db['synonyms'][element.upper()] = tmp['osiA'].upper()
        for element in tmp['siB']:
            check(loadBioGrid.db['synonyms'], element.upper())
            loadBioGrid.db['synonyms'][element.upper()] = tmp['osiB'].upper()
        if tmp['osiA'].upper() in loadBioGrid.db['synonyms']:
            tmp['osiA'] = loadBioGrid.db['synonyms'][tmp['osiA'].upper()]
        if tmp['osiB'].upper() in loadBioGrid.db['synonyms']:
            tmp['osiB'] = loadBioGrid.db['synonyms'][tmp['osiB'].upper()]
        """

        for element in comb:
            check(loadBioGrid.db, element[1].upper())
            loadBioGrid.db[element[1].upper()].add(element[0].upper())

        # check(loadBioGrid.db, tmp['osiA'].upper())
        # loadBioGrid.db[tmp['osiA'].upper()].add(tmp['osiB'].upper())
    p = pd.Series(loadBioGrid.db)
    p.to_hdf("BioGridPandas.h5", "biogrid")


def loadBioGridDict(fileName="BioGridPandas.h5"):
    if hasattr(loadBioGrid, "db"):
        return loadBioGrid.db
    if pd is None:
        print("pandas library not found.")
        loadBioGrid.db = {}
    else:
        try:
            loadBioGrid.db = pd.read_hdf(fileName, "biogrid")
        except:
            print("Biogrid database not found")
            loadBioGrid.db = {}

    return loadBioGrid.db


if __name__ == "__main__":
    # extractStatistics()
    db = loadBioGrid()
    # print len(db)
    # f = open('bioGridDict.dump', 'wb')
    # pickle.dump(db, f)
    # pass
    # db2 = loadBioGridDict()
    # print len(db2)
    # f = open('bioGridDict.dump', 'wb')
    # print len(db)
    # loadBioGrid()
    # print len(db2)



from bionetgen.modelapi.pattern import Molecule, Pattern
from bionetgen.modelapi.rulemod import RuleMod
from bionetgen.core.utils.utils import ActionList
from bionetgen.core.exc import BNGParseError


class ModelObj:
    """
    The base class for all items in a model (parameter, observable etc.).

    Attributes
    ----------
    comment : str
        comment at the end of the line/object
    line_label : str
        line label at the beginning of the line/object

    Methods
    -------
    print_line()
        generates the actual line string with line label and comments
        if applicable
    gen_string()
        generates the BNGL string of the object itself, separate from
        line attributes
    """

    def __init__(self):
        self._comment = None
        self._line_label = None

    def __str__(self) -> str:
        return self.gen_string()

    def __repr__(self) -> str:
        return self.gen_string()

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __delitem__(self, key):
        delattr(self, key)

    @property
    def comment(self) -> None:
        return self._comment

    @comment.setter
    def comment(self, val) -> None:
        # TODO: regex handling of # instead
        if val.startswith("#"):
            self._comment = val[1:]
        else:
            self._comment = val

    @property
    def line_label(self) -> str:
        return self._line_label

    @line_label.setter
    def line_label(self, val) -> None:
        # TODO: specific error handling
        try:
            ll = int(val)
            self._line_label = "{} ".format(ll)
        except:
            self._line_label = "{}: ".format(val)

    def print_line(self) -> str:
        s = "  "
        # let's deal with line label
        if self.line_label is not None:
            s += self.line_label
        # start building the rest of the string
        s += str(self)
        if self.comment is not None:
            s += " #{}".format(self.comment)
        return s


class Parameter(ModelObj):
    """
    Class for all parameters in the model, subclass of ModelObj.

    In BNGL parameters are of the form
        parameter_name parameter_value/expression
    or
        parameter_name = parameter_value/expression

    Attributes
    ----------
    name : str
        name of the parameter
    value : str
        value of the parameter, if loaded from XML this will always
        exist since NFsim needs the value and not the expression
    expr : str
        this exists if the parameter is a math expression, not necerssary
    write_expr : bool
        this is a boolean that determines if the generated string has
        is in expression form or in value form.
    """

    def __init__(self, name, value, expr=None):
        super().__init__()
        self.name = name
        self.value = value
        self.expr = expr
        try:
            test = float(expr)
            self.write_expr = False
        except:
            self.write_expr = True

    def gen_string(self) -> str:
        if self.write_expr:
            return "{} {}".format(self.name, self.expr)
        else:
            return "{} {}".format(self.name, self.value)


class Compartment(ModelObj):
    """
    Class for all compartments in the model, subclass of ModelObj.

    In BNGL the compartments are of the form
        compartment_name dimensions size
    or
        compartment_name dimensions size parent_compartment
    the second form only applies when one compartment is contained in
    another compartment.

    Attributes
    ----------
    name : str
        name of the compartment
    dim : str
        dimensionality of the compartment
    size : str
        size/volume of the compartment
    outside : str
        parent compartment, if exists
    write_expr : bool
        boolean that describes if the size is a volume or an expression
    """

    def __init__(self, name, dim, size, outside=None):
        super().__init__()
        self.name = name
        self.dim = dim
        self.size = size
        try:
            test = float(size)
            self.write_expr = False
        except:
            self.write_expr = True
        self.outside = outside

    def gen_string(self) -> str:
        s = "{} {} {}".format(self.name, self.dim, self.size)
        if self.outside is not None:
            s += " {}".format(self.outside)
        return s


class Observable(ModelObj):
    """
    Class for all observables in the model, subclass of ModelObj.

    In BNGL the observables are of the form
        observable_type observable_name observable_patterns
    where patterns can include multiple patterns separated by commas.

    Attributes
    ----------
    name : str
        name of the observable
    type : str
        type of the observable, Molecules or Species
    patterns : list[Pattern]
        list of patterns of the observable

    Methods
    -------
    add_pattern
        add a Pattern object into the list of patterns
        for this observable
    """

    def __init__(self, name, otype, patterns=[]):
        super().__init__()
        self.name = name
        self.type = otype
        if self.type == "Species":
            for pat in patterns:
                if pat.MatchOnce:
                    pat.MatchOnce = False
        self.patterns = patterns

    def gen_string(self) -> str:
        s = "{} {} ".format(self.type, self.name)
        for ipat, pat in enumerate(self.patterns):
            if ipat > 0:
                s += ","
            s += str(pat)
        return s

    def add_pattern(self, pat) -> None:
        # if type is species, set MatchOnce to false since all species automatically match once
        if self.type == "Species":
            pat.MatchOnce = False
        self.patterns.append(pat)


class MoleculeType(ModelObj):
    """
    Class for all molecule types in the model, subclass of ModelObj.

    In BNGL the molecule types are of the form
        molecule_type
    where all possible states of each component of a molecule is
    listed, e.g.
        A(b, p~0~1, k~ON~OFF~NULL)

    Attributes
    ----------
    molecule : Molecule
        a molecule type only contains a molecule object which
        can also handle multiple component states
    """

    def __init__(self, name, components):
        super().__init__()
        self.name = name
        self.molecule = Molecule(name=name, components=components)

    def gen_string(self) -> str:
        return str(self.molecule)


class Species(ModelObj):
    """
    Class for all species in the model, subclass of ModelObj.

    In BNGL the species/seed species are of the form
        species count
    where species is a single pattern and count is the starting
    value for that specific pattern

    Attributes
    ----------
    pattern : Pattern
        pattern of the seed species
    count : str
        starting value of the seed species
    """

    def __init__(self, pattern=Pattern(), count=0):
        super().__init__()
        self.pattern = pattern
        self.count = count
        self.name = str(self.pattern)

    def gen_string(self) -> str:
        s = "{} {}".format(self.pattern, self.count)
        return s


class Function(ModelObj):
    """
    Class for all functions in the model, subclass of ModelObj.

    In BNGL functions are of the form
        function_name function_expression
    or
        function_name = function_expression
    and functions can have arguments
        function_name(arg1, arg2, ..., argN)


    Attributes
    ----------
    name : str
        name of the function
    expr : str
        function expression
    args : list
        optional list of arguments for the function
    """

    def __init__(self, name, expr, args=None):
        super().__init__()
        self.name = name
        self.expr = expr
        self.args = args

    def gen_string(self) -> str:
        if self.args is None:
            s = "{} = {}".format(self.name, self.expr)
        else:
            s = "{}({}) = {}".format(self.name, ",".join(self.args), self.expr)
        return s


class Action(ModelObj):
    """
    Class for all actions in the model, subclass of ModelObj.

    In BNGL actions are of the form
        action_type({arg1=>value1, arg2=>value2, ...})

    Attributes
    ----------
    type : str
        type of action, e.g. simulate or writeFile
    args : dict[arg_name] = arg_value
        action arguments as keys and their values as values
    """

    def __init__(self, action_type=None, action_args={}) -> None:
        super().__init__()
        AList = ActionList()
        self.normal_types = AList.normal_types
        self.no_setter_syntax = AList.no_setter_syntax
        self.square_braces = AList.square_braces
        self.possible_types = AList.possible_types
        # Set initial values
        self.name = action_type
        self.type = action_type
        self.args = action_args
        # check type
        if self.type not in self.possible_types:
            raise BNGParseError(message=f"Action type {self.type} not recognized!")
        seen_args = []
        for arg in action_args:
            arg_name, arg_value = arg, action_args[arg]
            valid_arg_list = AList.arg_dict[self.type]
            # TODO: actions that don't take argument names should be parsed separately to check validity of arg-val tuples
            # TODO: currently not type checking arguments
            if valid_arg_list is None:
                raise BNGParseError(
                    message=f"Argument {arg_name} is given, but action {self.type} does not take arguments"
                )
            if len(valid_arg_list) > 0:
                if arg_name not in AList.arg_dict[self.type]:
                    raise BNGParseError(
                        message=f"Action argument {arg_name} not recognized!\nCheck to make sure action is correctly formatted"
                    )
                # TODO: If arg_value is the correct type
            if arg_name in seen_args:
                print(
                    f"Warning: argument {arg_name} already given, using latter value {arg_value}"
                )
            else:
                seen_args.append(arg_name)

    def gen_string(self) -> str:
        # TODO: figure out every argument that has special
        # requirements, e.g. method requires the value to
        # be a string
        action_str = "{}(".format(self.type)
        # we can skip curly if we don't have arguments
        # and we NEED to skip it for some actions
        if self.type in self.normal_types and not len(self.args) == 0:
            action_str += "{"
        elif self.type in self.square_braces:
            action_str += "["
        # add arguments
        for iarg, arg in enumerate(self.args):
            val = self.args[arg]
            if iarg > 0:
                action_str += ","
            # some actions need =>, some don't
            if self.type in self.normal_types:
                action_str += f"{arg}=>{val}"
            else:
                action_str += f"{arg}"
        # we can skip curly if we don't have arguments
        # and we NEED to skip it for some actions
        if self.type in self.normal_types and not len(self.args) == 0:
            action_str += "}"
        elif self.type in self.square_braces:
            action_str += "]"
        # close up the action
        action_str += ")"
        return action_str

    def print_line(self) -> str:
        s = ""
        # let's deal with line label
        if self.line_label is not None:
            s += self.line_label
        # start building the rest of the string
        s += str(self)
        if self.comment is not None:
            s += " #{}".format(self.comment)
        return s


class Rule(ModelObj):
    """
    Class for all rules in the model, subclass of ModelObj.

    Attributes
    ----------
    name : str
        name of the rule, optional
    reactants : list[Pattern]
        list of patterns for reactants
    products : list[Pattern]
        list of patterns for products
    rule_mod : RuleMod
        modifier (moveConnected, TotalRate, etc.) used by a given rule
    operations : list[Operation]
        list of operations

    Methods
    -------
    set_rate_constants((k_fwd,k_bck))
        sets forward and backwards rate constants, backwards rate
        constants are optional and if not given, will set the rule
        to be a unidirectional rule
    side_string(list[Pattern])
        given a list of patterns, return a string formatted to be
        on one side of a rule definition
    """

    def __init__(
        self,
        name,
        reactants=[],
        products=[],
        rate_constants=(),
        rule_mod=RuleMod(),
        operations=[],
    ) -> None:
        super().__init__()
        self.name = name
        self.reactants = reactants
        self.products = products
        if rule_mod is None:
            self.rule_mod = RuleMod()
        else:
            self.rule_mod = rule_mod
        self.operations = operations
        self.set_rate_constants(rate_constants)

    def set_rate_constants(self, rate_cts):
        if len(rate_cts) == 1:
            self.rate_constants = [rate_cts[0]]
            self.bidirectional = False
        elif len(rate_cts) == 2:
            self.rate_constants = [rate_cts[0], rate_cts[1]]
            self.bidirectional = True
        else:
            print("1 or 2 rate constants allowed")

    def gen_string(self):
        if self.bidirectional:
            return "{}: {} <-> {} {},{} {}".format(
                self.name,
                self.side_string(self.reactants),
                self.side_string(self.products),
                self.rate_constants[0],
                self.rate_constants[1],
                str(self.rule_mod),
            )
        else:
            return "{}: {} -> {} {} {}".format(
                self.name,
                self.side_string(self.reactants),
                self.side_string(self.products),
                self.rate_constants[0],
                str(self.rule_mod),
            )

    def side_string(self, patterns):
        side_str = ""
        for ipat, pat in enumerate(patterns):
            if ipat > 0:
                side_str += " + "
            side_str += str(pat)
        return side_str


class EnergyPattern(ModelObj):
    """
    Class for all energy patterns in the model, subclass of ModelObj.

    In BNGL the energy patterns are of the form
        EP_pattern EP_expression

    Attributes
    ----------
    name : str
        id of the energy pattern
    pattern : Pattern
        Pattern object representing the energy pattern
    expression : str
        expression used for energy pattern
    """

    def __init__(self, name, pattern, expression):
        super().__init__()
        self.name = name
        self.pattern = pattern
        self.expression = expression

    def gen_string(self) -> str:
        s = "{} {}".format(self.pattern, self.expression)
        return s


class PopulationMap(ModelObj):
    """
    Class for all population maps in the model, subclass of ModelObj.

    In BNGL the population maps are of the form
        structured_species -> population_species lumping_parameter

    Attributes
    ----------
    name : str
        id of the population map
    struct_species : Pattern
        Pattern object representing the species to be mapped
    pop_species : Pattern
        Pattern object representing the population count
    rate : str
        lumping parameter used in population mapping
    """

    def __init__(self, name, struct_species, pop_species, rate):
        super().__init__()
        self.name = name
        self.species = struct_species
        self.population = pop_species
        self.rate = rate

    def gen_string(self) -> str:
        s = "{} -> {} {}".format(self.species, self.population, self.rate)
        return s



from bionetgen.core.utils.logging import BNGLogger
from bionetgen.modelapi.pattern import Pattern, Molecule, Component
import pyparsing as pp


class BNGParsers:
    """
    Container for parsers for the reader
    """

    def __init__(self) -> None:
        pass


class BNGPatternReader:
    """
    Class that generates parsers to read BNG pattern strings and
    form Pattern objects from them.

    Usage: BNGPatternReader(pattern_string)

    Arguments
    ---------
    pattern_str : str
        The pattern string to read and generate a Pattern object from

    Attributes
    ----------
    pattern : Pattern
        The Pattern object formed from the parsed string
    parsers : BNGParsers
        Container object that has parsers for various parts of a
        BNG pattern string

    Methods
    -------
    define_parsers : None
        runs other defined parser commands to setup all parsers
    define_component_parser : None
        defines pyparsing parser for components
    define_molecule_parser : None
        defines pyparsing parser for molecules
    define_pattern_parser : None
        defines pyparsing parser for overall patterns
    make_pattern : Pattern
        forms the actual Pattern object from the pattern string
        using the defined parsers
    """

    def __init__(self, pattern_str) -> None:
        self.logger = BNGLogger()
        self.pattern_str = pattern_str
        self.parsers = BNGParsers()
        self.define_parsers()
        self.pattern = self.make_pattern(self.pattern_str)

    def define_parsers(self):
        self.define_component_parser()
        self.define_molecule_parser()
        self.define_pattern_parser()

    def define_component_parser(self):
        """
        Defines specific parsers for BNG components
        """
        # bng names are alpha numericals and _1
        self.parsers.base_name = pp.Word(pp.alphas, pp.alphanums + "_")
        # components have optional states and bonds
        self.parsers.state = pp.Combine(
            pp.Word("~") + (self.parsers.base_name ^ pp.Word(pp.nums))
        ) ^ pp.Word("~?")
        self.parsers.bond = pp.Combine(
            (pp.Word("!") + pp.Word(pp.nums)) ^ (pp.Word("!?")) ^ (pp.Word("!+"))
        )
        self.parsers.component = (
            self.parsers.base_name
            + pp.Optional(self.parsers.state)
            + pp.Optional(self.parsers.bond)
        )
        component_parser = pp.Combine(self.parsers.component)
        # components are separated by commas
        component_separator = pp.Word(",")
        self.parsers.components_parser = pp.delimited_list(
            component_parser, delim=component_separator
        )
        self.parsers.combined_components_parser = pp.delimited_list(
            component_parser, delim=component_separator, combine=True
        )

    def define_molecule_parser(self):
        """
        Defines specific parsers for BNG molecules
        """
        # molecules can have tags
        self.parsers.tag = pp.Combine(
            pp.Word("%") + (self.parsers.base_name ^ pp.Word(pp.nums))
        )
        # and compartments
        self.parsers.compartment = pp.Combine(pp.Word("@") + self.parsers.base_name)
        # combine tag and compartment
        tag_comp = (
            pp.Optional(self.parsers.tag) + pp.Optional(self.parsers.compartment)
        ) ^ (pp.Optional(self.parsers.compartment) + pp.Optional(self.parsers.tag))
        # full molecule
        self.parsers.molecule = (
            self.parsers.base_name
            + tag_comp
            + pp.Word("(")
            + pp.Optional(self.parsers.combined_components_parser)
            + pp.Word(")")
            + tag_comp
        )
        molecule_parser = pp.Combine(self.parsers.molecule)
        # molecules
        # components are separated by commas
        molecule_separator = pp.Word(".")
        self.parsers.molecules_parser = pp.delimited_list(
            molecule_parser, delim=molecule_separator
        )
        self.parsers.combined_molecules_parser = pp.delimited_list(
            molecule_parser, delim=molecule_separator, combine=True
        )

    def define_pattern_parser(self):
        """
        Defines specific parsers for overall BNG patterns
        """
        # a pattern can start with a tag or a compartment
        mods = pp.Word("$") ^ pp.Word("{MatchOnce}")
        # zero molecule is a simple 0
        zeroMolecule = pp.Word("0")
        # quantifier
        quantifier = pp.Combine(
            (
                pp.Word("<")
                ^ pp.Word("<=")
                ^ pp.Word("==")
                ^ pp.Word(">=")
                ^ pp.Word(">")
            )
            + pp.Word(pp.nums)
        )
        # combine tag and compartment
        tag = self.parsers.tag + (pp.Word(":") ^ pp.Word("::"))
        comp = self.parsers.compartment + (pp.Word(":") ^ pp.Word("::"))
        tag_comp_1 = (
            self.parsers.tag + self.parsers.compartment + (pp.Word(":") ^ pp.Word("::"))
        )
        tag_comp_2 = (
            self.parsers.compartment + self.parsers.tag + (pp.Word(":") ^ pp.Word("::"))
        )
        tag_comp = tag ^ comp ^ tag_comp_1 ^ tag_comp_2
        pattern = (
            pp.Optional(tag_comp)
            + pp.Optional(mods)
            + self.parsers.combined_molecules_parser
            + pp.Optional(quantifier)
        )
        # full pattern
        self.parsers.pattern = pattern ^ zeroMolecule

    def make_pattern(self, pattern_str):
        """
        Forms the Pattern object from the given string using the
        parsed defined above.
        """
        # if pattern_str == "X()":
        #     # import ipdb;ipdb.set_trace()
        #     import IPython,sys;IPython.embed();sys.exit()
        # set location for logging
        log_loc = f"{__file__} : BNGPatternReader.make_pattern()"
        # instantiate a pattern
        pattern = Pattern(molecules=[])
        # start parsing
        parsed_pattern = self.parsers.pattern.parseString(pattern_str)
        split_molecs = None
        # first we'll pull out any features that are pattern only
        for parsed_val in parsed_pattern:
            # these are pattern features and the entire molecules section
            if parsed_val.startswith("@"):
                # this is a pattern-wide compartment
                self.logger.debug(f"found compartment in {parsed_val}", loc=log_loc)
                pattern.compartment = parsed_val.replace("@", "")
                continue
            elif parsed_val.startswith("%"):
                # this is a pattern-wide tag
                self.logger.debug(f"found tag in {parsed_val}", loc=log_loc)
                pattern.label = parsed_val.replace("%", "")
                continue
            elif parsed_val.startswith(":"):
                # this is a pattern-wide separator
                self.logger.debug(f"found separator in {parsed_val}", loc=log_loc)
                continue
            elif ("$" in parsed_val) or ("{MatchOnce}" in parsed_val):
                # this is a constant value species pattern or a MatchOnce observable
                self.logger.debug(f"found mod in {parsed_val}", loc=log_loc)
                if "$" in parsed_val:
                    pattern.fixed = True
                elif "{MatchOnce}" in parsed_val:
                    pattern.MatchOnce = True
                continue
            elif (
                ("<" in parsed_val)
                or ("<=" in parsed_val)
                or ("==" in parsed_val)
                or (">=" in parsed_val)
                or (">" in parsed_val)
            ):
                self.logger.debug(f"found quantifier in {parsed_val}", loc=log_loc)
                if "==" in parsed_val:
                    pattern.relation = "=="
                elif "<=" in parsed_val:
                    pattern.relation = "<="
                elif ">=" in parsed_val:
                    pattern.relation = ">="
                elif "<" in parsed_val:
                    pattern.relation = "<"
                elif ">" in parsed_val:
                    pattern.relation = ">"
                pattern.quantity = int(parsed_val.replace(pattern.relation, ""))
                # this is a quantifier
                continue
            elif parsed_val == "0":
                # this is a zero molecule
                m = Molecule(components=[])
                m.parent_pattern = pattern
                pattern.molecules.append(m)
                self.logger.debug(f"found zero molecule in {parsed_val}", loc=log_loc)
                continue
            # only molecules should be remaining
            split_molecs = self.parsers.molecules_parser.parseString(parsed_val)
        # if we had a zero molecule we are done
        if split_molecs is None:
            # this is the zero molecule
            self.logger.debug(
                f"no molecules found in: {self.pattern_str}, done", loc=log_loc
            )
            return pattern
        # we got the molecule list, let's loop over molecules now
        self.logger.debug(f"molecules: {split_molecs}", loc=log_loc)
        for molec_str in split_molecs:
            molecule = Molecule(components=[])
            molecule.parent = pattern
            # each molec str is a molecule string with all features
            parsed_molec = self.parsers.molecule.parseString(molec_str)
            self.logger.debug(f"parsed molecule: {parsed_molec}", loc=log_loc)
            in_molec = False
            for parsed_val in parsed_molec:
                # we need to pull out the molecule features
                if parsed_val.startswith("@"):
                    # this is a molecule compartment
                    self.logger.debug(f"found compartment in {parsed_val}", loc=log_loc)
                    molecule.compartment = parsed_val.replace("@", "")
                    continue
                elif parsed_val.startswith("%"):
                    # this is a molecule tag
                    self.logger.debug(f"found tag in {parsed_val}", loc=log_loc)
                    molecule.label = parsed_val.replace("%", "")
                    continue
                elif parsed_val == "(" or parsed_val == ")":
                    if parsed_val == "(":
                        in_molec = True
                    else:
                        in_molec = False
                    # this molecule opening and closing
                    self.logger.debug(f"found paran in {parsed_val}", loc=log_loc)
                    continue
                if not in_molec:
                    # if we aren't in molecule yet, this can only be the name
                    molecule.name = parsed_val
                else:
                    # only components remain, parse those and loop
                    split_components = self.parsers.components_parser.parseString(
                        parsed_val
                    )
                    self.logger.debug(
                        f"split components: {split_components}", loc=log_loc
                    )
                    for component_str in split_components:
                        # each component is parsed separately now
                        parsed_component = self.parsers.component.parseString(
                            component_str
                        )

                        component = Component()
                        component.parent_molecule = molecule
                        # import ipdb;ipdb.set_trace()
                        for icomp, comp in enumerate(parsed_component):
                            if icomp == 0:
                                component.name = comp
                            elif "!" in comp:
                                # this is a bond, we'll have to figure out how to make the bonds
                                splt = comp.split("!")
                                component.bonds += list(
                                    filter(lambda x: len(x) > 0, splt)
                                )
                            elif "~" in comp:
                                splt = comp.split("~")
                                splt = list(filter(lambda x: len(x) > 0, splt))
                                if len(splt) > 1:
                                    component.states += splt
                                else:
                                    component.state = splt[0]
                            elif "%" in comp:
                                # this is a label, need to parse this
                                pass
                        # import IPython;IPython.embed()
                        # self._label = None
                        molecule.components.append(component)
                        self.logger.debug(
                            f"split components: {split_components}", loc=log_loc
                        )
            self.logger.debug(
                f"molecule parsed: {molecule}",
                loc=log_loc,
            )
            pattern.molecules.append(molecule)
        # ship the finalized pattern object
        pattern.canonicalize()
        return pattern



class NetworkObj:
    """
    The base class for all items in a network object (parameter, groups etc.).

    Attributes
    ----------
    comment : str
        comment at the end of the line/object
    line_label : str
        line label at the beginning of the line/object

    Methods
    -------
    print_line()
        generates the actual line string with line label and comments
        if applicable
    gen_string()
        generates the BNGL string of the object itself, separate from
        line attributes
    """

    def __init__(self):
        self._comment = ""
        self._line_label = None

    def __str__(self) -> str:
        return self.gen_string()

    def __repr__(self) -> str:
        return self.gen_string()

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __delitem__(self, key):
        delattr(self, key)

    @property
    def comment(self) -> None:
        return self._comment

    @comment.setter
    def comment(self, val) -> None:
        # TODO: regex handling of # instead
        if val is not None:
            if len(val) > 0:
                if val.startswith("#"):
                    self._comment = val[1:]
                else:
                    self._comment = val
            else:
                self._comment = None
        else:
            self._comment = None

    @property
    def line_label(self) -> str:
        return self._line_label

    @line_label.setter
    def line_label(self, val) -> None:
        # TODO: specific error handling
        try:
            ll = int(val)
            self._line_label = "{} ".format(ll)
        except:
            self._line_label = "{}: ".format(val)

    def print_line(self) -> str:
        s = "  "
        # let's deal with line label
        if self.line_label is not None:
            s += self.line_label
        # start building the rest of the string
        s += str(self)
        if self.comment is not None:
            s += " #{}".format(self.comment)
        return s


class NetworkParameter(NetworkObj):
    """
    Class for all parameters in the network, subclass of NetworkObj.

    In BNGL networks parameters are of the form
        parameter_ID parameter_name parameter_value/expression # comment

    Attributes
    ----------
    id : int
        ID/line label of the network parameter
    name : str
        name of the network parameter
    value : str
        value of the network parameter
    """

    def __init__(self, pid, name, value, comment=""):
        super().__init__()
        self.line_label = pid
        self.name = name
        self.value = value
        self.comment = comment

    def gen_string(self) -> str:
        s = "{} {}".format(self.name, self.value)
        return s


# TODO:
class NetworkCompartment(NetworkObj):
    """
    Class for all compartments in the network, subclass of NetworkObj.

    In BNGL the compartments are of the form
        compartment_name dimensions size

    Attributes
    ----------
    name : str
        name of the compartment
    dim : str
        dimensionality of the compartment
    size : str
        size/volume of the compartment
    outside : str
        parent compartment, if exists
    """

    def __init__(self, name, dim, size, outside=None):
        super().__init__()
        self.name = name
        self.dim = dim
        self.size = size
        self.outside = outside

    def gen_string(self) -> str:
        s = "{} {} {}".format(self.name, self.dim, self.size)
        if self.outside is not None:
            s += " {}".format(self.outside)
        return s


class NetworkGroup(NetworkObj):
    """
    Class for all groups in the network, subclass of NetworkObj.

    In BNGL networks the groups are of the form
        group_ID group_name group_species
    where species are separated by commas.

    Attributes
    ----------
    id : str
        id of the group
    name : str
        name of the group
    species : list[expr]
        list of species expressions of the group
    """

    def __init__(self, gid, name, members=[], comment=""):
        super().__init__()
        self.line_label = gid
        self.name = name
        self.members = members
        self.comment = comment

    def gen_string(self) -> str:
        s = "{} {} ".format(self.name, ",".join(self.members))
        return s


class NetworkSpecies(NetworkObj):
    """
    Class for all species in the network, subclass of NetworkObj.

    In BNGL the species/seed species are of the form
        species count
    where species is a single pattern and count is the starting
    value for that specific pattern

    Attributes
    ----------
    pattern : Pattern
        pattern of the seed species
    count : str
        starting value of the seed species
    """

    def __init__(self, sid, name, count=0, comment=""):
        super().__init__()
        self.line_label = sid
        self.name = name
        self.count = count
        self.comment = comment

    def gen_string(self) -> str:
        s = "{} {}".format(self.name, self.count)
        return s


# TODO:
class NetworkFunction(NetworkObj):
    """
    Class for all functions in the network, subclass of NetworkObj.

    In BNGL functions are of the form
        function_name function_expression

    Attributes
    ----------
    name : str
        name of the function
    expr : str
        function expression
    args : list
        optional list of arguments for the function
    """

    def __init__(self, name, expr, args=None):
        super().__init__()
        self.name = name
        self.expr = expr
        self.args = args

    def gen_string(self) -> str:
        if self.args is None:
            s = "{} = {}".format(self.name, self.expr)
        else:
            s = "{}({}) = {}".format(self.name, ",".join(self.args), self.expr)
        return s


# TODO:
class NetworkReaction(NetworkObj):
    """
    Class for all reactions in the network, subclass of NetworkObj.

    Attributes
    ----------
    name : str
        name of the rule, optional
    reactants : list[Pattern]
        list of patterns for reactants
    products : list[Pattern]
        list of patterns for products
    rule_mod : RuleMod
        modifier (moveConnected, TotalRate, etc.) used by a given rule
    operations : list[Operation]
        list of operations
    """

    def __init__(
        self,
        rid,
        reactants=[],
        products=[],
        rate_constant=None,
        comment=None,
    ) -> None:
        super().__init__()
        self.line_label = rid
        self.name = rid
        self.reactants = reactants
        self.products = products
        self.rate_constant = rate_constant
        self.comment = comment

    def gen_string(self):
        s = f"{','.join(self.reactants)} {','.join(self.products)} {self.rate_constant}"
        return s


# TODO:
class NetworkEnergyPattern(NetworkObj):
    """
    Class for all energy patterns in the network, subclass of NetworkObj.

    In BNGL the energy patterns are of the form
        EP_pattern EP_expression

    Attributes
    ----------
    name : str
        id of the energy pattern
    pattern : Pattern
        Pattern object representing the energy pattern
    expression : str
        expression used for energy pattern
    """

    def __init__(self, name, pattern, expression):
        super().__init__()
        self.name = name
        self.pattern = pattern
        self.expression = expression

    def gen_string(self) -> str:
        s = "{} {}".format(self.pattern, self.expression)
        return s


# TODO:
class NetworkPopulationMap(NetworkObj):
    """
    Class for all population maps in the model, subclass of ModelObj.

    In BNGL the population maps are of the form
        structured_species -> population_species lumping_parameter

    Attributes
    ----------
    name : str
        id of the population map
    struct_species : Pattern
        Pattern object representing the species to be mapped
    pop_species : Pattern
        Pattern object representing the population count
    rate : str
        lumping parameter used in population mapping
    """

    def __init__(self, name, struct_species, pop_species, rate):
        super().__init__()
        self.name = name
        self.species = struct_species
        self.population = pop_species
        self.rate = rate

    def gen_string(self) -> str:
        s = "{} -> {} {}".format(self.species, self.population, self.rate)
        return s



import cement
import bionetgen as bng
from cement.core.exc import CaughtSignal
from .core.exc import BNGError
from .core.exc import BNGVersionError
from .core.main import runCLI
from .core.main import plotDAT
from .core.main import runAtomizeTool
from .core.main import printInfo
from .core.main import visualizeModel
from .core.main import graphDiff
from .core.main import generate_notebook
from .core.utils.utils import test_perl

# pull defaults defined in core/defaults
CONF = bng.defaults
VERSION_BANNER = bng.defaults.banner

# require version argparse action
import argparse, sys
from pkg_resources import packaging


class requireAction(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)
        if values is not None:
            req_version = packaging.version.parse(values)
            cver = bng.core.version.get_version()
            cur_version = packaging.version.parse(cver)
            # if we don't meet requirement, warn user
            sys.tracebacklimit = 0
            if not (cur_version >= req_version):
                raise BNGVersionError(cur_version, req_version)
        # return super().__call__(parser, namespace, values, option_string=option_string)


class BNGBase(cement.Controller):
    """
    Base cement controller for BioNetGen CLI

    Used to set meta attributes like program name (label) as well
    as command line arguments. Each method is a subcommand in the
    command line with its own command line arguments.

    Subcommands
    -------
    run
        runs a model given by -i in folder given by -o
    notebook
        generates and opens a notebook for a model given by -i, optional
    plot
        plots a gdat/cdat/scan file given by -i into file supplied by -o
    info
        provides version and path information about the BNG installation and dependencies
    visualize
        provides various visualization options for BNG models
    """

    class Meta:
        label = "bionetgen"
        description = "A simple CLI to bionetgen <https://bionetgen.org>. Note that you need Perl installed."
        help = "bionetgen"
        arguments = [
            # TODO: Auto-load in BioNetGen version here
            (["-v", "--version"], dict(action="version", version=VERSION_BANNER)),
            # (['-s','--sedml'],dict(type=str,
            #                        default=CONF.config['bionetgen']['bngpath'],
            #                        help="Optional path to SED-ML file, if available the simulation \
            #                              protocol described in SED-ML will be ran")),
            (["-req", "--require"], dict(action=requireAction, type=str, default=None)),
            (
                ["-ll", "--log-level"],
                {
                    "help": 'This option allows you to select a logging level, from quietest to loudest options are: "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG". Default is set to INFO',
                    "default": None,
                    "type": str,
                },
            ),
        ]

    # This overwrites the default behavior and runs the CLI object from core/main
    # which in turn just calls BNG2.pl with the supplied options
    @cement.ex(
        help="Runs a given model using BNG2.pl",
        arguments=[
            (
                ["-i", "--input"],
                {
                    "help": "Path to BNGL file (required)",
                    "default": None,
                    "type": str,
                    "required": True,
                },
            ),
            (
                ["-o", "--output"],
                {
                    "help": 'Optional path to output folder (default: ".")',
                    "default": ".",
                    "type": str,
                },
            ),
            (
                ["-l", "--log"],
                {
                    "help": "saves BNG2.pl log to a file given (default: None)",
                    "default": None,
                    "type": str,
                    "dest": "log_file",
                },
            ),
            (
                ["--traceback-depth"],
                {
                    "help": "Sets the traceback depth for python. "
                    + "Defaults to 0 to avoid long tracebacks after a failed BNG2.pl call",
                    "default": 0,
                    "type": int,
                    "dest": "traceback_depth",
                },
            ),
        ],
    )
    def run(self):
        """
        This is the main run functionality of the CLI.

        It uses a convenience function defined in core/main
        to run BNG2.pl using subprocess, given the set of arguments
        in the command line and the configuraions set by the defaults
        as well as the end-user.
        """
        test_perl(app=self.app)
        runCLI(self.app)

    @cement.ex(
        help="Starts a Jupyter notebook to help run and analyze \
                  bionetgen models",
        arguments=[
            (
                ["-i", "--input"],
                {
                    "help": "Path to BNGL file to use with notebook",
                    "default": None,
                    "type": str,
                    "required": False,
                },
            ),
            (
                ["-o", "--output"],
                {
                    "help": "(optional) File to write the notebook in",
                    "default": "",
                    "type": str,
                },
            ),
            (
                ["-op", "--open"],
                {
                    "help": "(optional) If given, the notebook will be opened using nbopen",
                    "action": "store_true",
                },
            ),
        ],
    )
    def notebook(self):
        """
        Notebook subcommand that boots up a Jupyter notebook using the
        nbopen library. It uses a BNGNotebook class defined in core/notebook.

        The default template can be found under assets and in the future
        will likely be replaced by a standard templating tool (e.g. Jinja).

        The default base template is agnostic to the model and if -i is given
        the template then will be adjusted to load in the model supplied.
        """
        test_perl(app=self.app)
        generate_notebook(self.app)

    @cement.ex(
        help="Rudimentary plotting of gdat/cdat/scan files",
        arguments=[
            (
                ["-i", "--input"],
                {
                    "help": "Path to .gdat/.cdat file to use plot",
                    "default": None,
                    "type": str,
                    "required": True,
                },
            ),
            (
                ["-o", "--output"],
                {
                    "help": 'Optional path for the plot (default: "$model_name.png")',
                    "default": ".",
                    "type": str,
                },
            ),
            (
                ["--legend"],
                {
                    "help": "To plot the legend or not (default: False)",
                    "default": False,
                    "action": "store_true",
                    "required": False,
                },
            ),
            (
                ["--xmin"],
                {
                    "help": "x-axis minimum (default: determined from data)",
                    "default": None,
                    "type": float,
                },
            ),
            (
                ["--xmax"],
                {
                    "help": "x-axis maximum (default: determined from data)",
                    "default": False,
                    "type": float,
                },
            ),
            (
                ["--ymin"],
                {
                    "help": "y-axis minimum (default: determined from data)",
                    "default": False,
                    "type": float,
                },
            ),
            (
                ["--ymax"],
                {
                    "help": "y-axis maximum (default: determined from data)",
                    "default": False,
                    "type": float,
                },
            ),
            (["--xlabel"], {"help": "x-axis label (default: time)", "default": False}),
            (
                ["--ylabel"],
                {"help": "y-axis label (default: concentration)", "default": False},
            ),
            (
                ["--title"],
                {
                    "help": "title of plot (default: determined from input file)",
                    "default": False,
                },
            ),
        ],
    )
    def plot(self):
        """
        Plotting subcommand for very basic plotting using a convenience function
        defined in core/main.

        Currently we support gdat/cdat/scan file plotting, in a very basic manner.
        This command expects a space separated file where each column is a series.
        The first column is used for the x-axis and the rest is used as y-axis
        and every series is plotted.

        See bionetgen plot -h for all the allowed options.
        """
        plotDAT(self.app)

    @cement.ex(
        help="Provides version information for BNG and dependencies",
        arguments=[
            (
                ["-d", "--detail"],
                {
                    "help": "Adds more detail to the information printed.",
                    "default": False,
                    "action": "store_true",
                },
            ),
        ],
    )
    def info(self):
        """
        Information subcommand to provide installation versions and paths.

        Currently provides version information for BioNetGen, the BNG CLI, Perl,
        numpy, pandas, and libroadrunner. Also provides BNG2.pl and pyBNG paths.
        """
        printInfo(self.app)

    @cement.ex(
        help="Provides a simple way to get various visualizations of the model.",
        arguments=[
            (
                ["-i", "--input"],
                {
                    "help": "Path to BNGL model to visualize",
                    "default": None,
                    "type": str,
                    "required": True,
                },
            ),
            (
                ["-o", "--output"],
                {
                    "help": "(optional) Output folder, defaults to current folder",
                    "default": None,
                    "type": str,
                },
            ),
            (
                ["-t", "--type"],
                {
                    "help": "(optional) Type of visualization requested. Valid options are: "
                    + "'ruleviz_pattern','ruleviz_operation', 'contactmap', 'regulatory' and 'atom_rule'."
                    + " Regulatory and atom rule graphs are the same thing, defaults to 'contactmap'.",
                    "default": "",
                    "type": str,
                },
            ),
        ],
    )
    def visualize(self):
        """
        Subcommand to generate visualizations. Currently only supports visualize
        action from BioNetGen.

        Types of visualizations and their options
        - Rule pattern visualization: Visualization of each rule as a bipartite graph
        - Rule operation visualization: Visualization of each rule showing explicit graph operations
        - Contact map: Visualize the contact map of the model
        - Regulatory graph: Visualize the regulatory graph of the model, also called atom rule graph
        """
        test_perl(app=self.app)
        visualizeModel(self.app)

    @cement.ex(
        help="A subcommand to compare two contact maps made by BioNetGen. "
        + 'The default mode is "matrix" mode and will generate 4 graphs, two for each individual input '
        + "recolored, one for input 1 - input 2 and one for input 2 - input 1. This can be used to find "
        + "differences and communalities between two BNGL models of the same process. ",
        arguments=[
            (
                ["-i", "--input"],
                {
                    "help": "Path to the first graphml file to diff",
                    "default": None,
                    "type": str,
                    "required": True,
                },
            ),
            (
                ["-i2", "--input2"],
                {
                    "help": "Path to the second graphml file to diff",
                    "default": None,
                    "type": str,
                    "required": True,
                },
            ),
            (
                ["-o", "--output"],
                {
                    "help": "(optional) Output graphml file for the diff input 1 - input 2",
                    "default": None,
                    "type": str,
                    "required": False,
                },
            ),
            (
                ["-o2", "--output2"],
                {
                    "help": "(optional) Output graphml file for the diff input 2 - input 1",
                    "default": None,
                    "type": str,
                    "required": False,
                },
            ),
            (
                ["-m", "--mode"],
                {
                    "help": 'Diff mode. There are currently two available modes "matrix" and "union".',
                    "default": "matrix",
                    "type": str,
                },
            ),
            (
                ["-c", "--colors"],
                {
                    "help": "Path to the json file that contains color choices.",
                    "default": None,
                    "type": str,
                },
            ),
        ],
    )
    def graphdiff(self):
        # TODO: add documentation here
        """ """
        test_perl(app=self.app)
        graphDiff(self.app)

    @cement.ex(
        help="SBML to BNGL translator",
        arguments=[
            (
                ["-i", "--input"],
                {
                    "help": "input SBML file",
                    "default": None,
                    "type": str,
                    "required": True,
                },
            ),
            (
                ["-o", "--output"],
                {
                    "help": "output SBML file",
                    "default": ".",
                    "type": str,
                },
            ),
            (
                ["-t", "--annotation"],
                {
                    "help": "keep annotation information",
                    "default": False,
                    "action": "store_true",
                },
            ),
            (
                ["-c", "--convention-file"],
                {
                    "help": "Conventions file",
                    "type": str,
                },
            ),
            (
                ["-n", "--naming-conventions"],
                {
                    "help": "Naming conventions file",
                    "type": str,
                },
            ),
            (
                ["-u", "--user-structures"],
                {
                    "help": "User defined species",
                    "type": str,
                },
            ),
            (
                ["-id", "--molecule-id"],
                {
                    "help": "use SBML molecule ids instead of names. IDs are less descriptive but more bngl friendly. Use only if the generated BNGL has syntactic errors",
                    "default": False,
                    "action": "store_true",
                },
            ),
            (
                ["-a", "--atomize"],
                {
                    "help": "Infer molecular structure",
                    "default": False,
                    "action": "store_true",
                },
            ),
            (
                ["-p", "--pathwaycommons"],
                {
                    "help": "Don't use web services to infer information on the SBML species. Use this setting when you don't have an internet connection.",
                    "default": True,
                    "action": "store_false",
                },
            ),
            (
                ["-s", "--isomorphism-check"],
                {
                    "help": "disallow atomizations that produce the same graph structure",
                    "action": "store_true",
                },
            ),
            (
                ["-I", "--ignore"],
                {
                    "help": "ignore atomization translation errors",
                    "action": "store_true",
                },
            ),
            (
                ["-mr", "--memoized-resolver"],
                {
                    "help": "sometimes the dependency graph is too large and might cause a very large memory requirement. This option will slow the translator down but will decrease memory usage",
                    "default": False,
                    "action": "store_true",
                },
            ),
            (
                ["-k", "--keep-local-parameters"],
                {
                    "help": "this option will keep the local parameters unresolved so that they can be controlled from the parameter section in the BNGL. Without this option, local parameters will be resolved to their values in functions",
                    "default": False,
                    "action": "store_true",
                },
            ),
            (
                ["-q", "--quiet-mode"],
                {
                    "help": "this option will supress logging into STDIO and instead will write the logging into a file",
                    "default": False,
                    "action": "store_true",
                },
            ),
            (
                ["-omf", "--obs-map-file"],
                {
                    "help": "With this option enabled atomizer will print the map between SBML and BNGL observables in JSON format",
                    "default": None,
                    "type": str,
                },
            ),
            # (
            #     ["-cu", "--convert-units"],
            #     {
            #         "help": "convert units. Otherwise units are copied straight from sbml to bngl",
            #         "default": True,
            #         "action": "store_false",
            #     },
            # ),
        ],
    )
    def atomize(self):
        runAtomizeTool(self.app)


class BioNetGen(cement.App):
    """
    Cement app for BioNetGen CLI

    Used to set configuration options like config default,
    exiting on close and setting log handler. Currently set
    attributes are below.

    Attributes
    ----------
    label : str
        name of the application
    config_defaults : str
        the default set of configuration options, set in BNGDefaults object
    config_handler: str
        the name of the config handler, determines the syntax of the config files
    config_file_suffix: str
        the suffix to be used for config files
    config_files: list of str
        additional list of config files to enable
    exit_on_close : boolean
        determine if the app should exit when the key function is ran
    extensions : list of str
        extensions to be used with cement framework
    log_handler: str
        name of the log handler
    handlers: list of obj
        list of objects derived from cement.Controller that handles the actual CLI
    """

    class Meta:
        label = "bionetgen"

        # configuration defaults
        config_defaults = CONF.config

        # call sys.exit() on close
        exit_on_close = True

        # load additional framework extensions
        extensions = [
            "colorlog",
        ]

        # configuration handler
        config_handler = "configparser"

        # configuration file suffix
        config_file_suffix = ".conf"

        # add current folder to the list of config dirs
        config_files = ["./.{}.conf".format(label)]

        # set the log handler
        log_handler = "colorlog"

        # register handlers
        handlers = [BNGBase]


class BioNetGenTest(cement.TestApp, BioNetGen):
    """
    A sub-class of BioNetGen CLI application for testing
    purposes. See tests/test_bionetgen.py for examples.
    """

    class Meta:
        label = "bionetgen"


def main():
    with BioNetGen() as app:
        try:
            app.run()

        except AssertionError as e:
            print("AssertionError > %s" % e.args[0])
            app.exit_code = 1
            # TODO: figure out if this is what we want,
            # rn it prints stuff twice
            # if app.debug is True:
            #     import traceback

            #     traceback.print_exc()

        except BNGError as e:
            print("BNGError > %s" % e.args[0])
            app.exit_code = 1
            # TODO: figure out if this is what we want,
            # rn it prints stuff twice
            # if app.debug is True:
            #     import traceback

            #     traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print("\n%s" % e)
            app.exit_code = 0


if __name__ == "__main__":
    main()
